# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import pytest

from karabind import (
    AMPERE, IMAGEDATA_ELEMENT, INT32_ELEMENT, MANDATORY, METER, MICRO, MILLI,
    NDARRAY_ELEMENT, NODE_ELEMENT, OVERWRITE_ELEMENT, STRING_ELEMENT,
    AccessLevel, AccessType, ArchivePolicy, AssignmentType, DaqDataType, Hash,
    Logger, MetricPrefix, NodeType, Schema, Types, Unit, Validator,
    cppGraphicsRenderer1SchemaTest, cppShapeSchemaCircle,
    cppShapeSchemaEditableCircle, cppSomeClassSchemaSomeClassId,
    cppTestStruct1SchemaMyTest, cppTestStruct1SchemaTestStruct1, fullyEqual)
from karabo.common.states import State


def test_buildUp():
    try:
        schema = cppShapeSchemaCircle()
        assert schema.isAccessInitOnly("shadowEnabled")
        assert schema.isAccessInitOnly("radius")
        assert schema.isLeaf("radius")
        schema = cppShapeSchemaEditableCircle()
        allowedStates = schema.getOptions("state")
        assert allowedStates == ['INIT', 'ERROR', 'NORMAL']
        assert schema.getDefaultValue("state") == 'INIT'
        assert schema.getDefaultValue("stateN") == 'NORMAL'
        assert schema.getDefaultValue("stateE") == 'ERROR'
        assert schema.getDefaultValue("alarmW") == 'warn'
        assert schema.getDefaultValue("alarmA") == 'alarm'

        allowedStates = schema.getOptions("status")
        assert allowedStates == ['a', 'b', 'c']
        assert schema.getDefaultValue("status") == 'a'
    except Exception as e:
        pytest.fail(e, pytrace=True)

    try:
        schema = cppGraphicsRenderer1SchemaTest()
        assert schema.isAccessInitOnly("circle.radius")
        assert schema.isLeaf("circle.radius")
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getRootName():
    try:
        schema = cppTestStruct1SchemaMyTest()
        assert schema.getRootName() == "MyTest"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getTags():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getTags("exampleKey1")[0] == "hardware"
        assert schema.getTags("exampleKey1")[1] == "poll"
        assert schema.getTags("exampleKey2")[0] == "hardware"
        assert schema.getTags("exampleKey2")[1] == "poll"
        assert schema.getTags("exampleKey3")[0] == "hardware"
        assert schema.getTags("exampleKey3")[1] == "set"
        assert schema.getTags("exampleKey4")[0] == "software"
        assert schema.getTags("exampleKey5")[0] == "h/w"
        assert schema.getTags("exampleKey5")[1] == "d.m.y"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setTags():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.hasTags('x')
        assert schema.getTags('x') == ['IK', 'BH']
        schema.setTags('x', 'CY,SE')
        assert schema.getTags('x') == ['CY', 'SE']
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getsetExpertLevel():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        # getRequiredAccessLevel return access level as int
        assert schema.getRequiredAccessLevel('x') == AccessLevel.EXPERT
        assert schema.getRequiredAccessLevel('y') == AccessLevel.OPERATOR
        assert schema.getRequiredAccessLevel('a') == AccessLevel.OBSERVER

        schema.setRequiredAccessLevel('x', AccessLevel.EXPERT)
        schema.setRequiredAccessLevel('y', AccessLevel.OPERATOR)
        assert schema.getRequiredAccessLevel('x') == AccessLevel.EXPERT
        assert schema.getRequiredAccessLevel('y') == AccessLevel.OPERATOR
    except Exception as e:
        pytest.fail(e, pytrace=True)

    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getRequiredAccessLevel('exampleKey1') == \
            AccessLevel.OPERATOR
        assert schema.getRequiredAccessLevel('exampleKey2') == \
            AccessLevel.OPERATOR
        assert schema.getRequiredAccessLevel('exampleKey3') == \
            AccessLevel.EXPERT
        assert schema.getRequiredAccessLevel('exampleKey4') == \
            AccessLevel.EXPERT
        # default for readOnly
        assert schema.getRequiredAccessLevel('exampleKey5') == \
            AccessLevel.OBSERVER
        # default for reconfigurable
        assert schema.getRequiredAccessLevel('exampleKey10') == \
            AccessLevel.OPERATOR
        # observerAccess in reconfigurable
        assert schema.getRequiredAccessLevel('exampleKey11') == \
            AccessLevel.OBSERVER
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_isNode():
    schema = cppGraphicsRenderer1SchemaTest()
    assert schema.getRootName() == "test"
    assert schema.isNode("circle")
    assert schema.isNode("rectangle")
    assert not schema.isNode("shapes")


def test_getNodeType():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        nodeType = schema.getNodeType("exampleKey1")
        assert nodeType == NodeType.LEAF
        assert schema.getNodeType("exampleKey5") == NodeType.LEAF
        schema = cppGraphicsRenderer1SchemaTest()
        assert schema.getNodeType("shapes") == NodeType.LEAF
        assert schema.getNodeType("circle") == NodeType.NODE
        assert schema.getNodeType("rectangle") == NodeType.NODE
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getValueType():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getValueType("exampleKey1") == Types.STRING
        assert schema.getValueType("exampleKey2") == Types.INT32
        assert schema.getValueType("exampleKey3") == Types.UINT32
        assert schema.getValueType("exampleKey4") == Types.DOUBLE
        assert schema.getValueType("exampleKey5") == Types.INT64
        assert schema.getValueType("exampleKey7") == Types.VECTOR_INT32
        assert schema.getValueType("exampleKey8") == Types.VECTOR_DOUBLE
        assert schema.getValueType("exampleKey9") == Types.VECTOR_STRING
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getAliasAsString():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getAliasAsString("exampleKey2") == "10"
        assert schema.getAliasAsString("exampleKey3") == "5.5"

        assert schema.getAliasAsString("exampleKey4") == "exampleAlias4"
        assert schema.getAliasAsString("exampleKey5") == "exampleAlias5"
        assert schema.getAliasAsString("exampleKey6") == "1193046,43724"
        assert schema.getAliasAsString("testPath") == "5"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_keyHasAlias():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.keyHasAlias("exampleKey1") is False
        assert schema.keyHasAlias("exampleKey2")
        assert schema.keyHasAlias("exampleKey3")
        assert schema.keyHasAlias("exampleKey4")
        assert schema.keyHasAlias("exampleKey5")
        assert schema.keyHasAlias("exampleKey6")
        assert schema.keyHasAlias("testPath")
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_aliasHasKey():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.aliasHasKey(10)
        assert schema.aliasHasKey(5.5)
        assert schema.aliasHasKey("exampleAlias4")
        assert schema.aliasHasKey("exampleAlias5")
        assert schema.aliasHasKey([0x00123456, 0x0000aacc])
        assert schema.aliasHasKey(7) is False
        assert schema.aliasHasKey(5)
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getAliasFromKey():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getAliasFromKey("exampleKey2") == 10
        assert schema.getAliasFromKey("exampleKey3") == 5.5
        assert schema.getAliasFromKey("exampleKey4") == "exampleAlias4"
        assert schema.getAliasFromKey("exampleKey5") == "exampleAlias5"
        assert schema.getAliasFromKey("exampleKey6") == \
            [0x00123456, 0x0000aacc]
        assert schema.getAliasFromKey("testPath") == 5
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setAlias():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.getAliasFromKey("x") == 10
        schema.setAlias('x', 'abc')
        assert schema.getAliasFromKey("x") == 'abc'
        schema.setAlias('x', 99)
        assert schema.getAliasFromKey("x") == 99
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getKeyFromAlias():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getKeyFromAlias(10) == "exampleKey2"
        assert schema.getKeyFromAlias(5.5) == "exampleKey3"
        assert schema.getKeyFromAlias("exampleAlias4") == "exampleKey4"
        assert schema.getKeyFromAlias("exampleAlias5") == "exampleKey5"
        assert schema.getKeyFromAlias([0x00123456, 0x0000aacc]) == \
            "exampleKey6"
        assert schema.getKeyFromAlias(5) == "testPath"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getAccessMode():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getAccessMode("exampleKey1") == AccessType.WRITE
        assert schema.getAccessMode("exampleKey2") == AccessType.INIT
        assert schema.getAccessMode("exampleKey3") == AccessType.WRITE
        assert schema.getAccessMode("exampleKey4") == AccessType.INIT
        assert schema.getAccessMode("exampleKey5") == AccessType.READ
        assert schema.getAccessMode("exampleKey8") == AccessType.READ
        assert schema.getAccessMode("exampleKey10") == AccessType.WRITE
        assert schema.getAccessMode("testPath") == AccessType.WRITE
        assert schema.getAccessMode("testPath2") == AccessType.READ
        assert schema.getAccessMode("testPath3") == AccessType.INIT
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getAssignment():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getAssignment("exampleKey1") == AssignmentType.OPTIONAL
        assert schema.getAssignment("exampleKey2") == AssignmentType.OPTIONAL
        assert schema.getAssignment("exampleKey3") == AssignmentType.MANDATORY
        assert schema.getAssignment("exampleKey4") == AssignmentType.INTERNAL
        assert schema.getAssignment("exampleKey5") == AssignmentType.OPTIONAL
        assert schema.getAssignment("exampleKey8") == AssignmentType.OPTIONAL
        assert schema.getAssignment("exampleKey10") == AssignmentType.OPTIONAL
        assert schema.getAssignment("testPath") == AssignmentType.OPTIONAL
        assert schema.getAssignment("testPath2") == AssignmentType.OPTIONAL
        assert schema.getAssignment("testPath3") == AssignmentType.MANDATORY
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setAssignment():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.hasAssignment('x')
        assert schema.isAssignmentOptional('x')
        assert schema.isAssignmentMandatory('x') is False
        assert schema.getAssignment('x') == AssignmentType.OPTIONAL
        schema.setAssignment('x', AssignmentType.MANDATORY)
        assert schema.isAssignmentOptional('x') is False
        assert schema.isAssignmentMandatory('x')
        assert schema.getAssignment('x') == MANDATORY
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getDescription():
    schema = cppTestStruct1SchemaTestStruct1()
    assert schema.getDescription('exampleKey1') == "Example key 1 description"


def test_setDescription():
    schema = cppTestStruct1SchemaTestStruct1()
    schema.setDescription('exampleKey1', "No description")
    assert schema.getDescription('exampleKey1') == "No description"


def test_getOptions():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        options = schema.getOptions("exampleKey1")
        assert options[0] == "Radio"
        assert options[1] == "Air Condition"
        assert options[2] == "Navigation"

        assert schema.getOptions("exampleKey2")[0] == 5
        assert schema.getOptions("exampleKey2")[1] == 25
        assert schema.getOptions("exampleKey2")[2] == 10
        assert schema.getOptions("exampleKey4")[0] == 1.11
        assert schema.getOptions("exampleKey4")[1] == -2.22
        assert schema.getOptions("exampleKey4")[2] == 5.55

        assert schema.getOptions("testPath")[0] == "file1"
        assert schema.getOptions("testPath")[1] == "file2"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setOptions():
    try:
        schema = cppSomeClassSchemaSomeClassId()

        options = schema.getOptions("x")
        assert options[0] == 5
        assert options[1] == 25
        assert options[2] == 10
        assert schema.getOptions("x") == [5, 25, 10]

        schema.setOptions('x', '20, 5, 11, 13, 25')
        options = schema.getOptions("x")
        assert options == [20, 5, 11, 13, 25]
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_displayType():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        display = schema.getDisplayType("myNode")
        assert display == "WidgetNode"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_classId():
    try:
        schema = cppShapeSchemaCircle()
        assert schema.hasClassId("state")
        assert schema.getClassId("state") == "State"
        assert schema.hasClassId("alarm")
        assert schema.getClassId("alarm") == "AlarmCondition"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getDefaultValue():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getDefaultValue("exampleKey1") == "Navigation"
        assert schema.getDefaultValue("exampleKey2") == 10
        assert schema.getDefaultValueAs("exampleKey2", Types.STRING) == "10"
        assert schema.getDefaultValue("exampleKey5") == 1442244
        assert schema.getDefaultValueAs("exampleKey5", Types.STRING) == \
            "1442244"
        assert schema.getDefaultValue("exampleKey6") == 1.11
        assert schema.getDefaultValueAs("exampleKey6", Types.DOUBLE) == 1.11
        assert schema.getDefaultValue("exampleKey7") == [1, 2, 3]
        assert schema.getDefaultValue("exampleKey8") == [1.1, 2.2, 3.3]
        assert schema.getDefaultValue("exampleKey9") == ["Hallo", "World"]

        # 'readOnly'-element (vector as well) that does not specify
        #  'initialValue' has 'defaultValue' equal to string "" :
        assert schema.getDefaultValue("testPath2") == ""
        assert schema.getDefaultValue("vectInt") == []
        assert schema.getDefaultValue("exampleIntKey") == 20
        assert schema.getDefaultValue("exampleKey5") == 1442244
        assert schema.getDefaultValue("exampleKey7") == [1, 2, 3]

        # readOnly default specified by 'defaultValue'. not 'initialValue':
        assert schema.getDefaultValue("exampleKey5b") == 42
        assert schema.getDefaultValue("exampleKey7b") == [11, 22, 33]
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setDefaultValue():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.isAssignmentOptional('x')
        assert schema.hasDefaultValue('x')
        assert schema.getDefaultValue("x") == 5
        schema.setDefaultValue("x", 10)
        assert schema.getDefaultValue("x") == 10
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getAllowedStates():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        allowedStates = schema.getAllowedStates("exampleKey3")
        assert allowedStates[0] == State.STARTED
        assert allowedStates[1] == State.STOPPED
        assert allowedStates[2] == State.NORMAL
        assert schema.getAllowedStates("exampleKey3")[2] == State.NORMAL
        assert schema.getAllowedStates("exampleKey7")[0] == State.STARTED
        assert schema.getAllowedStates("exampleKey7")[1] == State.NORMAL
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getUnit():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getUnit("exampleKey2") == Unit.METER
        assert schema.getUnitName("exampleKey2") == "meter"
        assert schema.getUnitSymbol("exampleKey2") == "m"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setUnit():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.getUnit("x") == Unit.AMPERE
        schema.setUnit('x', METER)
        assert schema.getUnit("x") == METER
        assert schema.getUnit("x") == Unit.METER
        assert schema.getUnitName("x") == "meter"
        assert schema.getUnitSymbol("x") == "m"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getMetricPrefix():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getMetricPrefix("exampleKey2") == MetricPrefix.MILLI
        assert schema.getMetricPrefixName("exampleKey2") == "milli"
        assert schema.getMetricPrefixSymbol("exampleKey2") == "m"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setMetricPrefix():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.getMetricPrefix("x") == MetricPrefix.MILLI
        schema.setMetricPrefix("x", MetricPrefix.MICRO)
        assert schema.getMetricPrefix("x") == MICRO
        assert schema.getMetricPrefixName("x") == "micro"
        assert schema.getMetricPrefixSymbol("x") == "u"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getMinIncMaxInc():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getMinInc("exampleKey2") == 5
        assert schema.getMinIncAs("exampleKey2", Types.STRING) == "5"
        assert schema.getMaxInc("exampleKey2") == 25
        assert schema.getMaxIncAs("exampleKey2", Types.STRING) == "25"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setMinIncMaxInc():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.getMinInc("x") == 5
        assert schema.getMinIncAs("x", Types.STRING) == "5"
        assert schema.getMaxInc("x") == 25
        assert schema.getMaxIncAs("x", Types.STRING) == "25"
        schema.setMinInc('x', 3)
        schema.setMaxInc('x', 30)
        assert schema.getMinInc("x") == 3
        assert schema.getMinIncAs("x", Types.STRING) == "3"
        assert schema.getMaxInc("x") == 30
        assert schema.getMaxIncAs("x", Types.STRING) == "30"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_getMinExcMaxExc():
    schema = cppTestStruct1SchemaTestStruct1()
    try:
        assert schema.getMinExc("exampleKey3") == 10
        assert schema.getMinExc("exampleKey4") == -2.22
    except Exception as e:
        pytest.fail(e, pytrace=True)

    try:
        assert schema.getMaxExc("exampleKey3") == 20
        assert schema.getMaxExc("exampleKey4") == 5.55
    except Exception as e:
        pytest.fail(e, pytrace=True)

    try:
        assert schema.getMinExcAs("exampleKey3", Types.STRING) == "10"
        assert schema.getMinExcAs("exampleKey4", Types.STRING) == "-2.22"
    except Exception as e:
        pytest.fail(e, pytrace=True)

    try:
        assert schema.getMaxExcAs("exampleKey3", Types.STRING) == "20"
        assert schema.getMaxExcAs("exampleKey4", Types.STRING) == "5.55"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setMinExcMaxExc():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.getMinExc("y") == 0
        assert schema.getMaxExc("y") == 29
        schema.setMinExc("y", 2)
        schema.setMaxExc("y", 30)
        assert schema.getMinExc("y") == 2
        assert schema.getMaxExc("y") == 30
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_hasgetsetMinMaxSize():
    schema = cppTestStruct1SchemaTestStruct1()
    assert schema.hasMinSize('exampleKey10')
    assert schema.hasMaxSize('exampleKey10')
    assert schema.getMinSize('exampleKey10') == 2
    assert schema.getMaxSize('exampleKey10') == 7
    schema.setMinSize('exampleKey10', 5)
    schema.setMaxSize('exampleKey10', 12)
    assert schema.getMinSize('exampleKey10') == 5
    assert schema.getMaxSize('exampleKey10') == 12
    # ------
    assert schema.hasMinSize('exampleKey11') is False
    assert schema.hasMaxSize('exampleKey11') is False
    schema.setMinSize('exampleKey11', 1)
    schema.setMaxSize('exampleKey11', 42)
    assert schema.hasMinSize('exampleKey11')
    assert schema.hasMaxSize('exampleKey11')
    assert schema.getMinSize('exampleKey11') == 1
    assert schema.getMaxSize('exampleKey11') == 42


def test_vectorElement():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.isAccessReadOnly("exampleKey7")
        assert schema.hasDefaultValue("exampleKey7")
        assert schema.hasDefaultValue("exampleKey10")
        assert schema.hasDefaultValue("exampleKey11")
        assert schema.hasDefaultValue("exampleKey12")
        assert schema.hasDefaultValue("exampleKey14")
        assert schema.hasDefaultValue("exampleKey15")

        assert schema.getDefaultValue("exampleKey10") == [10, 20, 30]
        assert schema.getDefaultValue("exampleKey12") == [1.1, -2.2, 3.3]
        assert schema.getDefaultValue("exampleKey11") == [10, 20, 30]
        assert schema.getDefaultValueAs("exampleKey11", Types.STRING) == \
            "10,20,30"
        assert schema.getDefaultValueAs("exampleKey11",
                                        Types.VECTOR_INT32) == [10, 20, 30]
        assert schema.getDefaultValue("exampleKey14") == \
            ["Hallo", "World", "Test"]
        assert schema.getDefaultValueAs("exampleKey14", Types.STRING) == \
            "Hallo,World,Test"
        assert schema.getDefaultValue("exampleKey15") == \
            ["word1", "word2", "test"]
        assert schema.getDefaultValueAs("exampleKey15", Types.STRING) == \
            "word1,word2,test"
        assert schema.getMinSize("exampleKey10") == 2
        assert schema.getMaxSize("exampleKey10") == 7
    except Exception as e:
        pytest.fail(e, pytrace=True)

    try:
        schema = cppSomeClassSchemaSomeClassId()
        validator = Validator()
        configuration = Hash('somelist', [])
        ok, error, validated = validator.validate(schema, configuration)
        if not ok:
            raise RuntimeError(error)
        somelist = validated['somelist']
        somelist.append(99)
        somelist.append(55)
        configuration['somelist'] = somelist
        ok, error, validated = validator.validate(schema, configuration)
        if not ok:
            raise RuntimeError(error)
        assert validated is not None
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_ndarrayElement():
    schema = Schema()
    (
        NDARRAY_ELEMENT(schema).key("example16")
        .dtype("BOOL")
        .shape([2, 3])
        .unit(Unit.DEGREE_CELSIUS)
        .metricPrefix(MetricPrefix.CENTI)
        .commit(),

        NDARRAY_ELEMENT(schema).key("example17")
        .dtype("UINT32")
        .shape([2, 5, 0])
        .commit(),

        NDARRAY_ELEMENT(schema).key("example18")
        .dtype("FLOAT")
        .shape("3,2,1")
        .commit()
    )

    assert schema.getDefaultValue("example16.shape") == [2, 3]
    assert schema.getDefaultValue("example17.shape") == [2, 5, 0]
    assert schema.getDefaultValue("example18.shape") == [3, 2, 1]
    assert schema.getUnit("example16.data") == Unit.DEGREE_CELSIUS
    assert schema.getMetricPrefix("example16.data") == MetricPrefix.CENTI

    assert schema.isAccessReadOnly("example16")
    assert schema.isAccessReadOnly("example17")
    assert schema.isAccessReadOnly("example18")


# def test_listElement():
#     schema = \
#       Configurator(GraphicsRenderer2).getSchema("GraphicsRenderer2")
#     assert schema.getDisplayType("chars") == "unitTest"
#     assert schema.isAccessReconfigurable("chars") is True


def test_getDisplayType():
    schema = cppTestStruct1SchemaTestStruct1()
    assert schema.getDisplayType("exampleBitsKey1") == "bin"
    assert schema.getDisplayType("exampleBitsKey2") == \
        "bin|10:In Error, 21:Busy, 35:HV On, 55:Crate On"
    assert schema.getDisplayType("exampleBitsKey3") == "oct"
    assert schema.getDisplayType("exampleBitsKey4") == "hex"
    assert schema.getDisplayType("exampleKey5") == "Int64DisplayType"
    assert schema.getDisplayType("exampleKey8") == "Curve"
    assert schema.getDisplayType("exampleKey9") == "TestDisplayType"


def test_setDisplayType():
    try:
        schema = cppSomeClassSchemaSomeClassId()
        assert schema.hasDisplayType('y') is False
        schema.setDisplayType('y', 'blabla')
        assert schema.hasDisplayType('y')
        assert schema.getDisplayType("y") == "blabla"
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_isCommand():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.isCommand("slotTest")
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_isProperty():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.isProperty("slotTest") is False
        assert schema.isProperty("testPath2")
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_hasArchivePolicy():
    schema = cppTestStruct1SchemaTestStruct1()
    assert schema.hasArchivePolicy("exampleKey5") is False
    assert not schema.hasArchivePolicy("exampleKey6")
    assert not schema.hasArchivePolicy("exampleKey7")
    assert schema.hasArchivePolicy("exampleKey8")


def test_getArchivePolicy():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        assert schema.getArchivePolicy("exampleKey8") == \
            ArchivePolicy.NO_ARCHIVING
    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_setArchivePolicy():
    schema = cppSomeClassSchemaSomeClassId()
    assert not schema.hasArchivePolicy("a")
    schema.setArchivePolicy('a', ArchivePolicy.EVERY_EVENT)
    assert schema.getArchivePolicy("a") == ArchivePolicy.EVERY_EVENT
    schema.setArchivePolicy('a', ArchivePolicy.NO_ARCHIVING)
    assert schema.getArchivePolicy("a") == ArchivePolicy.NO_ARCHIVING


def test_perKeyFunctionality():
    try:
        schema = cppTestStruct1SchemaTestStruct1()
        keys = schema.getKeys()
        for key in keys:
            if key == "exampleKey1":
                assert schema.hasAssignment(key)
                assert schema.isAssignmentOptional(key)
                assert schema.hasDefaultValue(key)
                assert schema.hasAccessMode(key)
                assert schema.isAccessReconfigurable(key)
                assert schema.hasOptions(key)
                assert schema.hasTags(key)
                assert schema.hasUnit(key) is False
                assert schema.hasMetricPrefix(key) is False

            if key == "exampleKey2":
                assert schema.hasDefaultValue(key)
                assert schema.hasAccessMode(key)
                assert schema.isAccessInitOnly(key)
                assert schema.hasOptions(key)
                assert schema.hasTags(key)
                assert schema.hasAllowedStates(key) is False
                assert schema.hasUnit(key)
                assert schema.hasMetricPrefix(key)
                assert schema.hasMinInc(key)
                assert schema.hasMaxInc(key)

            if key == "exampleKey3":
                assert schema.hasAssignment(key)
                assert schema.isAssignmentMandatory(key)
                assert schema.hasDefaultValue(key) is False
                assert schema.hasOptions(key) is False
                assert schema.hasAllowedStates(key)
                assert schema.hasMinExc(key)
                assert schema.hasMaxExc(key)

            if key == "exampleKey4":
                assert schema.hasDefaultValue(key) is False
                assert schema.isAssignmentInternal(key)
                assert schema.hasAccessMode(key)
                assert schema.isAccessInitOnly(key)

            if key == "exampleKey5":
                assert schema.hasDefaultValue(key)
                assert schema.hasAssignment(key)
                assert schema.isAssignmentOptional(key)
                assert schema.hasAccessMode(key)
                assert schema.isAccessReadOnly(key)

    except Exception as e:
        pytest.fail(e, pytrace=True)


def test_merge():
    schema = Schema()
    (
        STRING_ELEMENT(schema).key("a")
        .description("a").displayedName("a")
        .assignmentOptional().defaultValue("a value")
        .tags("CY,CY,NC,JS,KW,NC")
        .commit(),

    )
    assert "a" in schema
    assert "x" not in schema
    assert "y" not in schema
    assert "z" not in schema

    schema2 = Schema()
    (
        STRING_ELEMENT(schema2).key("x")
        .description("x")
        .displayedName("x")
        .assignmentOptional().defaultValue("a value")
        .tags("LM,BH")
        .commit(),

        STRING_ELEMENT(schema2).key("y")
        .tags("CY")
        .displayedName("Example key 1")
        .description("Example key 1 description")
        .options("Radio,Air Condition,Navigation", ",")
        .assignmentOptional().defaultValue("Radio")
        .reconfigurable()
        .commit(),

        INT32_ELEMENT(schema2).key("z").alias(10)
        .tags("CY,LM,KW")
        .displayedName("Example key 2")
        .description("Example key 2 description")
        .options("5, 25, 10")
        .minInc(5).maxInc(25).unit(AMPERE).metricPrefix(MILLI)
        .assignmentOptional().defaultValue(10)
        .init()
        .commit(),
    )
    assert "x" in schema2
    assert "y" in schema2
    assert "z" in schema2

    schema += schema2

    assert "a" in schema
    assert "x" in schema
    assert "y" in schema
    assert "z" in schema


def test_logger():
    config = Hash("level", "DEBUG")
    Logger.configure(config)
    root = Logger.getLogger()
    a1 = Logger.getLogger("a1")
    a1_a2 = Logger.getLogger("a1.a2")
    root.debug("ERROR")

    Logger.useConsole()
    root.debug("ok")
    a1.debug("ok")
    a1_a2.debug("ok")
    root.INFO("ok")
    a1.INFO("ok")
    a1_a2.INFO("ok")

    Logger.reset()
    root.debug("ERROR")
    a1.debug("ERROR")
    a1_a2.debug("ERROR")


def test_helpFunction():
    pass
    # uncomment to see help:
    # schema = TestStruct1.getSchema("TestStruct1")
    # schema.help()


def test_schemaImageElement():
    schema = Schema()
    (
        IMAGEDATA_ELEMENT(schema).key("myImage1")
        .displayedName("myImage")
        .description("Image1 Element")
        .setDimensions("100, 200")
        .operatorAccess()
        .commit(),

        IMAGEDATA_ELEMENT(schema).key("myImage2")
        .displayedName("myImage2")
        .description("Image2 Element")
        .setDimensions([110, 210])
        .operatorAccess()
        .commit(),
    )
    assert schema.getDisplayType("myImage1") == "ImageData"
    assert schema.isCustomNode("myImage1") is True
    assert schema.getCustomNodeClass("myImage1") == "ImageData"
    assert schema.getAccessMode("myImage1") == AccessType.READ
    assert schema.getNodeType("myImage1") == NodeType.NODE
    assert schema.getRequiredAccessLevel("myImage1") == AccessLevel.OPERATOR
    assert schema.getDisplayedName("myImage1") == "myImage"
    assert schema.getDescription("myImage1") == "Image1 Element"
    assert schema.getDefaultValue("myImage1.dims") == [100, 200]
    assert schema.getDefaultValue("myImage2.dims") == [110, 210]

    ide = IMAGEDATA_ELEMENT(schema).key("moreImageElement")
    with pytest.raises(RuntimeError):
        ide.setDimensions(123)  # only list ans str are allowed

    assert schema.getDescription("myImage1.pixels") == \
        "The N-dimensional array containing the pixels"
    assert schema.getValueType("myImage1.pixels.data") == Types.BYTE_ARRAY
    assert schema.getDisplayedName("myImage1.dims") == "Dimensions"
    assert schema.getValueType("myImage1.dims") == Types.VECTOR_UINT64
    assert schema.getDisplayType("myImage1.dims") == "Curve"
    assert schema.getDisplayedName("myImage1.encoding") == "Encoding"
    assert schema.getValueType("myImage1.encoding") == Types.INT32

    assert schema.getDisplayedName("myImage1.pixels.isBigEndian") == \
        "Is big-endian"
    assert schema.getValueType("myImage1.pixels.isBigEndian") == Types.BOOL
    assert schema.getDefaultValue("myImage1.pixels.isBigEndian") is False


def test_allowed_states():
    # The 'states' argument is tuple
    schema = cppTestStruct1SchemaTestStruct1()
    schema.setAllowedStates("exampleKey3", (State.ACQUIRING, State.NORMAL,
                            State.PASSIVE))
    allowedStates = schema.getAllowedStates("exampleKey3")
    assert allowedStates == [State.ACQUIRING, State.NORMAL, State.PASSIVE]
    # The 'states' argument is list
    schema = cppTestStruct1SchemaTestStruct1()
    schema.setAllowedStates("exampleKey3", [State.ACQUIRING, State.NORMAL,
                                            State.PASSIVE])
    allowedStates = schema.getAllowedStates("exampleKey3")
    assert allowedStates == [State.ACQUIRING, State.NORMAL, State.PASSIVE]
    # The 'states' argument is a 'set'
    schema = cppTestStruct1SchemaTestStruct1()
    schema.setAllowedStates("exampleKey3", {State.ACQUIRING, State.NORMAL,
                                            State.PASSIVE})
    allowedStates = schema.getAllowedStates("exampleKey3")
    # the insertion order is not guaranteed for python 'set' ...
    assert State.ACQUIRING in allowedStates
    assert State.NORMAL in allowedStates
    assert State.PASSIVE in allowedStates
    # The 'states' are *args
    schema = cppTestStruct1SchemaTestStruct1()
    schema.setAllowedStates("exampleKey3", State.ACQUIRING, State.NORMAL,
                            State.PASSIVE)
    allowedStates = schema.getAllowedStates("exampleKey3")
    assert allowedStates == [State.ACQUIRING, State.NORMAL, State.PASSIVE]


def test_daq_data_type():
    schema = Schema()

    NODE_ELEMENT(schema).key(
        "trainData").commit()  # Has no DAQ data type yet
    NODE_ELEMENT(schema).key("pulseData").setDaqDataType(
        DaqDataType.PULSE).commit()

    assert schema.hasDaqDataType("trainData") is False
    assert schema.hasDaqDataType("pulseData") is True
    assert schema.getDaqDataType("pulseData") == DaqDataType.PULSE

    # Now add DAQ data type to node "trainData"
    schema.setDaqDataType("trainData", DaqDataType.TRAIN)
    assert schema.hasDaqDataType("trainData")
    assert schema.getDaqDataType("trainData") == DaqDataType.TRAIN


def test_table_element():
    schema = cppTestStruct1SchemaTestStruct1()

    # Both, defaultValue and initialValue should work for read-only tables
    default = schema.getDefaultValue("tableI")
    assert len(default) == 1
    assert fullyEqual(default[0], Hash("int", 2))
    default = schema.getDefaultValue("tableD")
    assert len(default) == 1
    assert fullyEqual(default[0], Hash("int", 3))

    # There are more tests of tables in the integration tests...


def test_overwrite_restrictions_for_options():
    schema = Schema()
    (
        INT32_ELEMENT(schema).key("range")
        .displayedName("Range")
        .options("0,1")
        .assignmentOptional().defaultValue(0)
        .reconfigurable()
        .commit()
    )
    vec = schema.getOptions("range")
    assert len(vec) == 2
    assert vec[0] == 0
    assert vec[1] == 1
    (
        OVERWRITE_ELEMENT(schema).key("range")
        .setNewOptions("0,1,2")
        .commit()
    )
    vec = schema.getOptions("range")
    assert len(vec) == 3
    assert vec[0] == 0
    assert vec[1] == 1
    assert vec[2] == 2


def test_overwrite_tags():
    schema = Schema()
    (
        INT32_ELEMENT(schema).key("taggedProp")
        .tags("bip, bop")
        .assignmentOptional().defaultValue(0)
        .reconfigurable()
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 2
    assert vec[0] == "bip"
    assert vec[1] == "bop"
    (
        OVERWRITE_ELEMENT(schema).key("taggedProp")
        .setNewTags("doff")
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 1
    assert vec[0] == "doff"
    (
        OVERWRITE_ELEMENT(schema).key("taggedProp")
        .setNewTags(["doff", "deff"])
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 2
    assert vec[0] == "doff"
    assert vec[1] == "deff"
    (
        OVERWRITE_ELEMENT(schema).key("taggedProp")
        .setNewTags(("chip", "chop"))
        .commit()
    )
    vec = schema.getTags("taggedProp")
    assert len(vec) == 2
    assert vec[0] == "chip"
    assert vec[1] == "chop"
    # Only iterable of str are allowed
    with pytest.raises(RuntimeError):
        (
            OVERWRITE_ELEMENT(schema).key("taggedProp")
            .setNewTags((1, 2))
            .commit()
        )



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import Hash, Types


def test_attributes_1():
    h = Hash("a.b.a.b", 42)
    h.setAttribute("a.b.a.b", "attr1", "someValue")
    assert h.getAttribute("a.b.a.b", "attr1") == "someValue"


def test_attributes_2():
    h = Hash("a.b.a.b", 42)
    h.setAttribute("a.b.a.b", "attr1", "someValue")
    assert h.getAttribute("a.b.a.b", "attr1") == "someValue"

    h.setAttribute("a.b.a.b", "attr2", 42)
    assert h.getAttribute("a.b.a.b", "attr1") == "someValue"
    assert h.getAttribute("a.b.a.b", "attr2") == 42

    h.setAttribute("a.b.a.b", "attr2", 43)
    assert h.getAttribute("a.b.a.b", "attr1") == "someValue"
    assert h.getAttribute("a.b.a.b", "attr2") == 43

    h.setAttribute("a.b.a.b", "attr1", True)
    assert h.getAttribute("a.b.a.b", "attr1")
    assert h.getAttribute("a.b.a.b", "attr2") == 43

    attrs = h.getAttributes("a.b.a.b")
    assert attrs.size() == 2
    assert attrs.get("attr1")
    assert attrs.get("attr2") == 43

    node = attrs.getNode("attr1")
    assert node.getType() == Types.BOOL

    node = attrs.getNode("attr2")
    assert node.getType() == Types.INT32


def test_attributes_3():
    h = Hash("a.b.a.b", 1)
    h.setAttribute("a.b.a.b", "attr1", [1, 2, 3])
    assert h.getAttribute("a.b.a.b", "attr1")[1], 2


def test_attributes_4():
    h = Hash("a.b.a.b", 42)
    h.setAttribute("a.b.a.b", "attr1", "someValue")
    assert h.getAttribute("a.b.a.b", "attr1") == "someValue"


def test_attributes_5():
    h = Hash("a.b.a.b", 42)
    h.setAttribute("a.b.a.b", "attr1", "someValue")
    assert h.getAttribute("a.b.a.b", "attr1") == "someValue"



#!/usr/bin/env python

# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

__author__ = "gero.flucke@xfel.eu"
__date__ = "September, 2017, 13:45 PM"

import threading
import time

import numpy as np

from karabo import __version__ as karaboVersion
from karabo.bound import (
    BOOL_ELEMENT, DOUBLE_ELEMENT, FLOAT_ELEMENT, IMAGEDATA_ELEMENT,
    INPUT_CHANNEL, INT32_ELEMENT, INT64_ELEMENT, KARABO_CLASSINFO,
    NDARRAY_ELEMENT, NODE_ELEMENT, OUTPUT_CHANNEL, OVERWRITE_ELEMENT,
    SLOT_ELEMENT, STRING_ELEMENT, TABLE_ELEMENT, UINT32_ELEMENT,
    UINT64_ELEMENT, VECTOR_BOOL_ELEMENT, VECTOR_CHAR_ELEMENT,
    VECTOR_DOUBLE_ELEMENT, VECTOR_FLOAT_ELEMENT, VECTOR_INT32_ELEMENT,
    VECTOR_INT64_ELEMENT, VECTOR_STRING_ELEMENT, VECTOR_UINT32_ELEMENT,
    VECTOR_UINT64_ELEMENT, AlarmCondition, DaqDataType, Encoding, Hash,
    ImageData, MetricPrefix, PythonDevice, Schema, State, Types, Unit)


@KARABO_CLASSINFO("PropertyTest", karaboVersion)
class PropertyTest(PythonDevice):

    defVectorMaxSize = 10  # maximum size of pipeline vector

    def __init__(self, configuration):
        # always call PythonDevice constructor first!
        super().__init__(configuration)

        # Define the slots
        self.KARABO_SLOT(self.setAlarm)
        self.KARABO_SLOT(self.writeOutput)
        self.KARABO_SLOT(self.startWritingOutput)
        self.KARABO_SLOT(self.stopWritingOutput)
        self.KARABO_SLOT(self.resetChannelCounters)
        self.KARABO_SLOT(self.eosOutput)
        self.KARABO_SLOT(self.node_increment)
        self.KARABO_SLOT(self.node_reset)

        self._writingOutput = False
        self._writingMutex = threading.Lock()
        self._writingWorker = None

        # Define first function to be called after the constructor has finished
        self.registerInitialFunction(self.initialization)

    @staticmethod
    def expectedParameters(expected):
        '''Description of device parameters statically known'''
        (
            OVERWRITE_ELEMENT(expected).key("state")
            .setNewOptions(State.INIT, State.NORMAL, State.ERROR,
                           State.STARTED, State.STOPPING)
            .setNewDefaultValue(State.INIT)
            .commit(),

            BOOL_ELEMENT(expected).key("boolProperty")
            .displayedName("Bool")
            .description("A bool property")
            .reconfigurable()
            .assignmentOptional().defaultValue(False)
            .commit(),

            BOOL_ELEMENT(expected).key("boolPropertyReadOnly")
            .displayedName("Readonly Bool")
            .description("A bool property for testing alarms")
            .readOnly().initialValue(True)
            .commit(),

            # CHAR_ELEMENT, (U)INT8_ELEMENT, (U)INT16_ELEMENT do not exists

            INT32_ELEMENT(expected).key("int32Property")
            .displayedName("Int32")
            .description("An int32 property")
            .reconfigurable()
            .assignmentOptional().defaultValue(-32_000_000)
            .commit(),

            INT32_ELEMENT(expected).key("int32PropertyReadOnly")
            .displayedName("Readonly Int32")
            .description("An int32 property for testing")
            .readOnly().initialValue(32_000_000)
            .commit(),

            UINT32_ELEMENT(expected).key("uint32Property")
            .displayedName("UInt32")
            .description("A uint32 property")
            .reconfigurable()
            .assignmentOptional().defaultValue(32_000_000)
            .commit(),

            UINT32_ELEMENT(expected).key("uint32PropertyReadOnly")
            .displayedName("Readonly UInt32")
            .description("A uint32 property for testing")
            .readOnly().initialValue(32_000_000)
            .commit(),

            INT64_ELEMENT(expected).key("int64Property")
            .displayedName("Int64")
            .description("An int64 property")
            .reconfigurable()
            .assignmentOptional().defaultValue(3_200_000_000)
            .commit(),

            INT64_ELEMENT(expected).key("int64PropertyReadOnly")
            .displayedName("Readonly Int64")
            .description("An int64 property for testing")
            .readOnly().initialValue(3_200_000_000)
            .commit(),

            UINT64_ELEMENT(expected).key("uint64Property")
            .displayedName("UInt64 property")
            .description("A uint64 property")
            .reconfigurable()
            .assignmentOptional().defaultValue(3_200_000_000)
            .commit(),

            UINT64_ELEMENT(expected).key("uint64PropertyReadOnly")
            .displayedName("Readonly UInt64")
            .description("A uint64 property for testing")
            .readOnly().initialValue(3_200_000_000)
            .commit(),

            FLOAT_ELEMENT(expected).key("floatProperty")
            .displayedName("Float property")
            .description("A float property")
            .reconfigurable()
            .assignmentOptional().defaultValue(3.141596)
            .commit(),

            FLOAT_ELEMENT(expected).key("floatPropertyReadOnly")
            .displayedName("Readonly Float")
            .description("A float property for testing")
            .readOnly().initialValue(3.141596)
            .commit(),

            DOUBLE_ELEMENT(expected).key("doubleProperty")
            .displayedName("Double")
            .description("A double property")
            .reconfigurable()
            .assignmentOptional().defaultValue(0.)
            .commit(),

            DOUBLE_ELEMENT(expected).key("doublePropertyReadOnly")
            .displayedName("Readonly Double")
            .description("A double property for testing")
            .readOnly().initialValue(0.)
            .commit(),

            STRING_ELEMENT(expected).key("stringProperty")
            .displayedName("String")
            .description("A string property")
            .reconfigurable()
            .assignmentOptional().defaultValue("Some arbitrary text.")
            .commit(),

            SLOT_ELEMENT(expected).key("setAlarm")
            .displayedName("Set Alarm")
            .description("Set an acknowledgment requiring alarm to value of "
                         "String property - if convertable")
            .commit(),

            NODE_ELEMENT(expected).key("vectors")
            .displayedName("Vectors")
            .description("A node containing vector properties")
            .commit(),

            VECTOR_BOOL_ELEMENT(expected).key("vectors.boolProperty")
            .displayedName("Bool property")
            .description("A vector boolean property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([True, False, True,
                                                False, True, False])
            .commit(),

            VECTOR_CHAR_ELEMENT(expected).key("vectors.charProperty")
            .displayedName("Char property")
            .description("A vector character property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue(['A', 'B', 'C', 'D', 'E', 'F'])
            .commit(),

            VECTOR_INT32_ELEMENT(expected).key("vectors.int32Property")
            .displayedName("Int32 property")
            .description("A vector int32 property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([20000041, 20000042, 20000043,
                                                20000044, 20000045, 20000046])
            .commit(),

            VECTOR_UINT32_ELEMENT(expected).key("vectors.uint32Property")
            .displayedName("UInt32 property")
            .description("A vector uint32 property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([90000041, 90000042, 90000043,
                                                90000044, 90000045, 90000046])
            .commit(),

            VECTOR_INT64_ELEMENT(expected).key("vectors.int64Property")
            .displayedName("Int64 property")
            .description("A vector int64 property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([20000000041, 20000000042,
                                                20000000043, 20000000044,
                                                20000000045, 20000000046])
            .commit(),

            VECTOR_UINT64_ELEMENT(expected).key("vectors.uint64Property")
            .displayedName("UInt64 property")
            .description("A vector uint64 property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([90000000041, 90000000042,
                                                90000000043, 90000000044,
                                                90000000045, 90000000046])
            .commit(),

            VECTOR_FLOAT_ELEMENT(expected).key("vectors.floatProperty")
            .displayedName("Float property")
            .description("A vector float property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([1.23456, 2.34567, 3.45678,
                                                4.56789, 5.67891, 6.78912])
            .commit(),

            VECTOR_DOUBLE_ELEMENT(expected).key("vectors.doubleProperty")
            .displayedName("Double property")
            .description("A vector double property")
            .reconfigurable()
            .minSize(1)
            .maxSize(10)
            .assignmentOptional().defaultValue([1.234567891, 2.345678912,
                                                3.456789123, 4.567891234,
                                                5.678901234, 6.123456789])
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("vectors.stringProperty")
            .displayedName("String property")
            .description("A vector string property")
            .reconfigurable()
            .minSize(1).maxSize(10)
            .assignmentOptional().defaultValue(["AAAAA", "BBBBB", "CCCCC",
                                                "XXXXX", "YYYYY", "ZZZZZ"])
            .commit(),
        )

        # Now prepare table
        columns = Schema()
        (
            STRING_ELEMENT(columns).key("e1")
            .displayedName("E1")
            .description("E1 property")
            .assignmentOptional().defaultValue("E1")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(columns).key("e2")
            .displayedName("E2")
            .description("E2 property")
            .reconfigurable()
            .assignmentOptional().defaultValue(False)
            .commit(),

            INT32_ELEMENT(columns).key("e3")
            .displayedName("E3")
            .description("E3 property")
            .reconfigurable()
            .assignmentOptional().defaultValue(42)
            .commit(),

            FLOAT_ELEMENT(columns).key("e4")
            .displayedName("E4")
            .description("E4 property")
            .assignmentOptional().defaultValue(-3.14)
            .reconfigurable()
            .commit(),

            DOUBLE_ELEMENT(columns).key("e5")
            .displayedName("E5")
            .description("E5 property")
            .assignmentOptional().defaultValue(3.14)
            .reconfigurable()
            .commit(),
        )

        (

            TABLE_ELEMENT(expected).key("table")
            .displayedName("Table property")
            .description("Table containing one node.")
            .setColumns(columns)
            .assignmentOptional().defaultValue([Hash("e1", "abc", "e2", True,
                                                     "e3", 12, "e4", 0.9837,
                                                     "e5", 1.2345),
                                                Hash("e1", "xyz", "e2", False,
                                                     "e3", 42, "e4", 2.33333,
                                                     "e5", 7.77777)])
            .reconfigurable()
            .commit(),
        )
        # Transform access mode of columns for readOnly table
        for key in columns.getKeys():
            OVERWRITE_ELEMENT(columns).key(key).setNowReadOnly().commit()
        (

            TABLE_ELEMENT(expected).key("tableReadOnly")
            .displayedName("Read-only table property")
            .description("Read-only table with two values.")
            .setColumns(columns)
            .readOnly().initialValue([Hash("e1", "abc", "e2", True,
                                           "e3", 12, "e4", 0.9837,
                                           "e5", 1.2345),
                                      Hash("e1", "xyz", "e2", False,
                                           "e3", 42, "e4", 2.33333,
                                           "e5", 7.77777)])
            .commit(),
        )

        pipeData = Schema()
        (
            NODE_ELEMENT(pipeData).key("node")
            .displayedName("Node for DAQ")
            .description("An intermediate node needed by DAQ")
            .setDaqDataType(DaqDataType.TRAIN)
            .commit(),

            INT32_ELEMENT(pipeData).key("node.int32")
            .description("A signed 32-bit integer sent via the pipeline")
            .readOnly()
            .commit(),

            STRING_ELEMENT(pipeData).key("node.string")
            .description("A string send via the pipeline")
            .readOnly()
            .commit(),

            VECTOR_INT64_ELEMENT(pipeData).key("node.vecInt64")
            .description("A vector of signed 64-bit integers sent via the "
                         "pipeline")
            .maxSize(PropertyTest.defVectorMaxSize)  # DAQ needs that
            .readOnly()
            .commit(),

            NDARRAY_ELEMENT(pipeData).key("node.ndarray")
            .description("A multi dimensional array of floats sent via the "
                         "pipeline")
            .dtype(Types.FLOAT)
            .shape("100,200")
            .commit(),

            IMAGEDATA_ELEMENT(pipeData).key("node.image")
            .description("An image with pixels as 16-bit unsigned integers "
                         "sent via the pipeline")
            .setDimensions("400,500")
            .setEncoding(Encoding.GRAY)
            .setType(Types.UINT16)
            .commit(),
        )

        (
            OUTPUT_CHANNEL(expected).key("output")
            .displayedName("Output")
            .dataSchema(pipeData)
            .commit(),

            SLOT_ELEMENT(expected).key("writeOutput")
            .displayedName("Write to Output")
            .description("Write once to output channel 'Output'")
            .allowedStates(State.NORMAL)
            .commit(),

            FLOAT_ELEMENT(expected).key("outputFrequency")
            .displayedName("Output frequency")
            .description("The target frequency for continously writing to "
                         "'Output'")
            .unit(Unit.HERTZ)
            .maxInc(1000)
            .minExc(0.0)
            .assignmentOptional().defaultValue(1.0)
            .expertAccess()
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("outputCounter")
            .displayedName("Output Counter")
            .description("Last value sent as 'int32' via output channel "
                         "'Output'")
            .readOnly()
            .initialValue(0)
            .commit(),

            SLOT_ELEMENT(expected).key("startWritingOutput")
            .displayedName("Start Writing")
            .description("Start writing continously to output channel "
                         "'Output'")
            .allowedStates(State.NORMAL)
            .commit(),

            SLOT_ELEMENT(expected).key("stopWritingOutput")
            .displayedName("Stop Writing")
            .description("Stop writing continously to output channel "
                         "'Output'")
            .allowedStates(State.STARTED)
            .commit(),

            SLOT_ELEMENT(expected).key("eosOutput")
            .displayedName("EOS to Output")
            .description("Write end-of-stream to output channel 'Output'")
            .allowedStates(State.NORMAL)
            .commit(),

            INPUT_CHANNEL(expected).key("input")
            .displayedName("Input")
            .commit(),

            UINT32_ELEMENT(expected).key("processingTime")
            .displayedName("Processing Time")
            .description("Processing time of input channel data handler")
            .assignmentOptional()
            .defaultValue(0)
            .reconfigurable()
            .unit(Unit.SECOND)
            .metricPrefix(MetricPrefix.MILLI)
            .expertAccess()
            .commit(),

            INT32_ELEMENT(expected).key("currentInputId")
            .displayedName("Current Input Id")
            .description("Last value received as 'int32' on input channel "
                         "(default: 0)")
            .readOnly().initialValue(0)
            .commit(),

            UINT32_ELEMENT(expected).key("inputCounter")
            .displayedName("Input Counter")
            .description("Number of data items received on input channel")
            .readOnly().initialValue(0)
            .commit(),

            UINT32_ELEMENT(expected).key("inputCounterAtEos")
            .displayedName("Input Counter @ EOS")
            .description("Value of 'Input Counter' when endOfStream was "
                         "received")
            .readOnly().initialValue(0)
            .commit(),

            SLOT_ELEMENT(expected).key("resetChannelCounters")
            .displayedName("Reset Channels")
            .description("Reset counters involved in input/output channel "
                         "data flow")
            .allowedStates(State.NORMAL)
            .commit(),


            NODE_ELEMENT(expected).key("node")
            .displayedName("Node")
            .commit(),

            SLOT_ELEMENT(expected).key("node.increment")
            .displayedName("Increment")
            .description("Increment 'Counter read-only'")
            .commit(),

            SLOT_ELEMENT(expected).key("node.reset")
            .displayedName("Reset")
            .description("Reset 'Counter read-only'")
            .commit(),

            UINT32_ELEMENT(expected).key("node.counterReadOnly")
            .displayedName("Counter read-only")
            .readOnly()
            .initialValue(0)
            .commit(),

            UINT32_ELEMENT(expected).key("node.counter")
            .displayedName("Counter")
            .description("Values will be transferred to 'Counter read-only' "
                         "under same node")
            .reconfigurable()
            .assignmentOptional()
            .defaultValue(0)
            .commit(),
        )

        (
            STRING_ELEMENT(expected)
            .key("stringInternal")
            .displayedName("Internal String")
            .description("A string property with assignment internal")
            .reconfigurable()
            .assignmentInternal()
            .defaultValue("Internal Only")
            .commit(),

            STRING_ELEMENT(expected)
            .key("stringInitInternal")
            .displayedName("Internal Init String")
            .description("A string property with assignment internal "
                         "and init only")
            .init()
            .assignmentInternal()
            .defaultValue("Init Internal Only")
            .commit(),
        )

    def initialization(self):
        self.updateState(State.NORMAL)
        self.KARABO_ON_DATA("input", self.onData)
        self.KARABO_ON_EOS("input", self.onEndOfStream)

    def preReconfigure(self, incomingCfg):
        props = ["boolProperty", "int32Property", "uint32Property",
                 "int64Property", "uint64Property",
                 "floatProperty", "doubleProperty",
                 "table", "node.counter"]

        bulkSets = Hash()
        for prop in props:
            if prop in incomingCfg:
                readOnlyProp = prop + "ReadOnly"
                bulkSets[readOnlyProp] = incomingCfg[prop]
        if bulkSets:
            self.set(bulkSets)

    def setAlarm(self):
        alarm = AlarmCondition.fromString(self["stringProperty"])
        self.setAlarmCondition(alarm)

    def writeOutput(self):
        outputCounter = self["outputCounter"]
        outputCounter += 1
        self.set("outputCounter", outputCounter)

        # Set all numbers inside to outputCounter:
        data = Hash("node", Hash())
        node = data["node"]

        # setAs needed if counter exceeds INT32 range...
        node.setAs("int32", outputCounter, Types.INT32)
        node.set("string", str(outputCounter))
        # Using plain Hash.set(..), type of vector is determined from 1st elem:
        node.setAs("vecInt64", outputCounter * self.defVectorMaxSize,
                   Types.VECTOR_INT64)
        arr = np.full((100, 200), outputCounter, dtype=np.float32)
        node.set("ndarray", arr)
        imArr = np.full((400, 500), outputCounter, dtype=np.uint16)
        node.set("image", ImageData(imArr, encoding=Encoding.GRAY))

        self.writeChannel("output", data)

    def eosOutput(self):
        self.signalEndOfStream("output")

    def startWritingOutput(self):
        self._writingOutput = True
        self.updateState(State.STARTED)
        self._writingWorker = threading.Thread(target=self.writeLoop)
        self._writingWorker.start()

    def stopWritingOutput(self):
        self.updateState(State.STOPPING)
        with self._writingMutex:
            self._writingOutput = False
        self._writingWorker.join()
        self._writingWorker = None

    def resetChannelCounters(self):
        self.set(Hash("inputCounter", 0,
                      "inputCounterAtEos", 0,
                      "outputCounter", 0,
                      "currentInputId", 0))

    def writeLoop(self):
        # Always run at least once: write should start immediately
        shouldWrite = True
        while shouldWrite:
            before = time.time()
            self.writeOutput()

            with self._writingMutex:
                shouldWrite = self._writingOutput

            if shouldWrite:
                # Waits for an interval as close as possible to the interval
                # defined by the nominal outputFrequency.
                delay = 1.0 / self.get("outputFrequency")
                delay = max(0., delay - (time.time() - before))
                time.sleep(delay)

                # "Refreshes" the shouldWrite flag before entering a new loop
                # interaction During the sleep an stop command might have been
                # issued and we want to avoid the extra writeOutput call.
                with self._writingMutex:
                    shouldWrite = self._writingOutput

        self.updateState(State.NORMAL)

    def onData(self, data, meta):
        # Sleeps to simulate heavy work.
        procTimeSecs = self["processingTime"]/1000.0
        time.sleep(procTimeSecs)

        currentInputId = data.get("node.int32")
        inputCounter = self["inputCounter"]

        self.set(Hash("currentInputId", currentInputId,
                      "inputCounter", inputCounter+1))

        # Writes data received to output channel to allow Property_Test to
        # build pipelines of chained devices.
        self.writeOutput()

    def onEndOfStream(self, inputChannel):
        inputCounter = self["inputCounter"]
        self.set("inputCounterAtEos", inputCounter)
        # Forward endOfStream as well
        self.signalEndOfStream("output")

    def node_increment(self):
        counter = self.get("node.counterReadOnly")
        self.set("node.counterReadOnly", counter + 1)

    def node_reset(self):
        self.set("node.counterReadOnly", 0)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import time
import uuid
from collections.abc import Iterable
from threading import Thread

from karabind import (
    DeviceClient, EventLoop, Hash, Logger, SignalSlotable, startDeviceServer,
    stopDeviceServer)


def test_device_client_sync_api():
    # Run CPP event loop in background ...
    loopThread = Thread(target=EventLoop.work)
    loopThread.start()

    # Configure and start device server ...
    # ... and make sure that plugins contains DataGenerator
    serverId = f'cppServer/{uuid.uuid4()}'
    config = Hash("serverId", serverId)
    startDeviceServer(config)
    # Create DeviceClient ...
    clientId = str(uuid.uuid4())
    c = DeviceClient(clientId)
    assert c.getInstanceId() == clientId
    assert serverId in c.getServers()
    # instantiate 'data/prop/1' device ... timeout = 5secs
    # synopsis: instantiate(serverId, classId, config, timeout)
    deviceId = f'data/prop/{uuid.uuid4()}'
    c.instantiate(
        serverId, 'PropertyTest', Hash('deviceId', deviceId), 5)
    # Check that device was created ...
    assert deviceId in c.getDevices(serverId)

    assert c.get(deviceId, 'node.counter') == 0
    c.set(deviceId, 'node.counter', 10)
    assert c.get(deviceId, 'node.counter') == 10
    assert c.get(deviceId, 'node.counterReadOnly') == 10
    # increment node.counter ...
    for i in range(3):
        c.execute(deviceId, 'node.increment', 5)

    assert c.get(deviceId, 'node.counter') == 13
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    device_schema = c.getDeviceSchema(deviceId)
    assert 'outputFrequency' in device_schema.getParameterHash()
    assert 'output.connections' in device_schema.getParameterHash()

    active_schema = c.getActiveSchema(deviceId)
    assert 'int16Property' in active_schema.getParameterHash()
    assert 'table' in active_schema.getParameterHash()

    class_schema = c.getClassSchema(serverId, 'PropertyTest')
    assert 'karaboVersion' in class_schema.getParameterHash()
    assert 'boolProperty' in class_schema.getParameterHash()

    props = c.getProperties(deviceId)
    assert 'status' in props
    assert 'vectors.boolProperty' in props

    comms = c.getCurrentlyExecutableCommands(deviceId)
    assert 'writeOutput' in comms
    assert 'eosOutput' in comms

    props = c.getCurrentlySettableProperties(deviceId)
    assert 'int32Property' in props
    assert 'boolProperty' in props

    names = c.getOutputChannelNames(deviceId)
    assert isinstance(names, Iterable)  # it's a 'list', but let's not assert
    assert 'output' in names

    schema = c.getOutputChannelSchema(deviceId, 'output')
    assert 'node.ndarray.shape' in schema
    assert 'node.image.pixels.data' in schema
    assert 'node.image.pixels.shape' in schema

    # shutdown device ...
    c.killDevice(deviceId, 5)
    assert deviceId not in c.getDevices(serverId)
    # stop and de-register device server...
    stopDeviceServer(serverId)
    Logger.reset()
    EventLoop.stop()
    loopThread.join()


def test_device_client_async_api():
    # Run CPP event loop in background ...
    loopThread = Thread(target=EventLoop.work)
    loopThread.start()

    # Configure and start device server ...
    serverId = f'cppServer/{uuid.uuid4()}'
    config = Hash("serverId", serverId)
    startDeviceServer(config)
    # Create DeviceClient ...
    clientId = str(uuid.uuid4())
    c = DeviceClient(clientId)
    assert c.getInstanceId() == clientId
    assert serverId in c.getServers()

    # instantiate (no wait!) 'data/prop/<uuid>' device ... timeout = 5secs
    # synopsis: instantiateNoWait(serverId, classId, config)
    deviceId = f'data/prop/{uuid.uuid4()}'
    c.instantiateNoWait(serverId, 'PropertyTest', Hash('deviceId', deviceId))
    trials = 1000
    while trials > 0 and deviceId not in c.getDevices(serverId):
        time.sleep(0.002)
        trials -= 1
    assert deviceId in c.getDevices(serverId)

    assert c.get(deviceId, 'node.counter') == 0

    c.setNoWait(deviceId, 'node.counter', 10)
    trials = 1000
    while trials > 0 and c.get(deviceId, 'node.counter') != 10:
        time.sleep(0.02)
        trials -= 1
    assert c.get(deviceId, 'node.counter') == 10
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    # increment counter
    for i in range(3):
        c.executeNoWait(deviceId, 'node.increment')
    trials = 1000
    while trials > 0 and c.get(deviceId, 'node.counter') != 13:
        time.sleep(0.02)
        trials -= 1

    assert c.get(deviceId, 'node.counter') == 13
    assert c.get(deviceId, 'node.counterReadOnly') == 10

    # shutdown device ...
    c.killDeviceNoWait(deviceId)
    trials = 1000
    while trials > 0 and deviceId in c.getDevices(serverId):
        time.sleep(0.002)
        trials -= 1
    assert deviceId not in c.getDevices(serverId)

    # stop and de-register device server...
    stopDeviceServer(serverId)
    Logger.reset()
    EventLoop.stop()
    loopThread.join()


def test_slots_with_args():

    # Run CPP event loop in background ...
    loopThread = Thread(target=EventLoop.work)
    loopThread.start()

    deviceId = "slotWithArgsTester"
    sigSlot = SignalSlotable(deviceId, Hash(), 60, Hash())

    product = 0  # Updated by argSlotMultiplyInternal

    def argSlotMultiplyInternal(num1, num2):
        """A slot with args that returns no value"""
        nonlocal product
        product = num1 * num2

    def argSlotMultiply(num1, num2):
        """A slot with args that returns a single value"""
        nonlocal sigSlot
        sigSlot.reply(num1 * num2)

    def argSlotDivide(dividend, divisor):
        """A slot with args that returns two values"""
        nonlocal sigSlot
        sigSlot.reply(dividend // divisor, dividend % divisor)

    def argSlotThreeCases(input_str):
        """A slot with arg that returns three values"""
        nonlocal sigSlot
        sigSlot.reply(input_str.upper(), input_str.lower(), input_str)

    def argSlotZahlen(digit):
        """A slot with arg that returns four values"""
        nonlocal sigSlot
        digit_dict = {
            "ptBR": ["Zero", "Um", "Dois", "Três", "Quatro", "Cinco", "Seis",
                     "Sete", "Oito", "Nove"],
            "enUS": ["Zero", "One", "Two", "Three", "Four", "Five", "Six",
                     "Seven", "Eight", "Nine"],
            "deDE": ["Null", "Eins", "Zwei", "Drei", "Vier", "Fünf", "Sechs",
                     "Sieben", "Acht", "Neun"],
            "roman": ["? (Romans didn't have zero)", "I", "II", "III", "IV",
                      "V", "VI", "VII", "VIII", "IX"]
        }
        idx = digit % 10
        sigSlot.reply(digit_dict["ptBR"][idx], digit_dict["enUS"][idx],
                      digit_dict["deDE"][idx], digit_dict["roman"][idx])

    sigSlot.start()
    cli = DeviceClient("DeviceClientUnderTest")

    # Slot with two args and no return value (actually returns an empty tuple)
    sigSlot.registerSlot(argSlotMultiplyInternal)
    emptyTuple = cli.executeN(deviceId, "argSlotMultiplyInternal", 12, 4)
    assert product == 48
    assert len(emptyTuple) == 0

    # Slot with two args and one return value
    sigSlot.registerSlot(argSlotMultiply)
    multResult, = cli.executeN(deviceId, "argSlotMultiply", 12, 4)
    assert multResult == 48

    # Slot with two args and two return values
    sigSlot.registerSlot(argSlotDivide)
    (quotient, remainder) = cli.executeN(deviceId, "argSlotDivide", 12, 4)
    assert quotient == 3
    assert remainder == 0

    # Slot with one arg and three return values
    sigSlot.registerSlot(argSlotThreeCases)
    (allUp, allLow, echoed) = cli.executeN(deviceId,
                                           "argSlotThreeCases", "Input_Str",
                                           timeoutInSeconds=20)
    assert allUp == "INPUT_STR"
    assert allLow == "input_str"
    assert echoed == "Input_Str"

    # Slot with one arg and four return values
    sigSlot.registerSlot(argSlotZahlen)
    (ptBr, enUS, deDE, roman) = cli.executeN(
        deviceId, "argSlotZahlen", 5, timeoutInSeconds=15)
    assert ptBr == "Cinco"
    assert enUS == "Five"
    assert deDE == "Fünf"
    assert roman == "V"

    EventLoop.stop()
    loopThread.join()



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

import threading
import time
import uuid
from contextlib import nullcontext as does_not_raise
from functools import partial

import pytest

from karabo.bound import EventLoop, Hash, SignalSlotable


@pytest.fixture(scope="module")
def eventLoopFixt():
    loop_thread = threading.Thread(target=EventLoop.work)
    loop_thread.start()

    yield  # now all (since scope="module") tests are executed

    EventLoop.stop()
    loop_thread.join(timeout=10)
    assert not loop_thread.is_alive()


class SignalSlotableWithSlots(SignalSlotable):
    def slotOneArg(self, a):
        pass

    def slotVarArgs(self, *args):
        self.reply(len(args))

    def slotTwoArgDef(self, a, b, c=42):
        self.reply(a + b + c)

    # A keyword only argument with default
    def slotKwargs(self, *args, a=77):
        self.reply(len(args), a)


def test_sigslot_register_function(eventLoopFixt):

    timeout = 2000  # timeout in ms

    sigSlot = SignalSlotableWithSlots("sigSlot" + str(uuid.uuid4()))
    sigSlot.start()

    def funcOneArg(a):
        pass

    def funcVarArgs(*args):
        sigSlot.reply(len(args))

    def funcTwoArgDef(a, b, c=42):
        sigSlot.reply(a + b + c)

    # A keyword only argument with default
    def funcKwargs(*args, a=77):
        sigSlot.reply(len(args), a)

    # Ordinary registration - register under <function>.__name__
    sigSlot.registerSlot(funcOneArg)
    # "" => self messaging!
    sigSlot.request("", "funcOneArg", 1).waitForReply(timeout)

    # Register under defined name (and test '/' as part of name)
    sigSlot.registerSlot(funcOneArg, "func/oneArg1")
    req = sigSlot.request("", "func/oneArg1", "no_matter")
    req.waitForReply(timeout)

    # Register function with variable number of arguments.
    # Automatic number of argument detection gives zero:
    sigSlot.registerSlot(funcVarArgs)
    req = sigSlot.request("", "funcVarArgs")
    (nArg, ) = req.waitForReply(timeout)
    assert nArg == 0
    # NOTE: We require STRICT match for number of arguments.
    #       It's in underlying C++ code as well.
    req = sigSlot.request("", "funcVarArgs", 1)
    with pytest.raises(RuntimeError) as excinfo:
        req.waitForReply(timeout)
        assert "mismatched number of args" in str(excinfo.value)

    # We can register '*args' function with fixed number
    # (Also test that '_' can be requested as '.' is needed
    #  for nested slots in Schema)
    sigSlot.registerSlot(funcVarArgs, "func_varArgs1", 1)
    req = sigSlot.request("", "func.varArgs1", "no_matter2")
    (nArg, ) = req.waitForReply(timeout)
    assert nArg == 1

    # Then we must not miss to sent argument
    req = sigSlot.request("", "func_varArgs1")  # here access with '_'
    with pytest.raises(RuntimeError):
        req.waitForReply(timeout)

    #
    # Register function with default argument.
    #
    # Automatic number of argument detection includes the default
    # (cannot change that for backward compatibility...).
    sigSlot.registerSlot(funcTwoArgDef, "funcTwoArgDef3")
    req = sigSlot.request("", "funcTwoArgDef3", 1, 2, 3)
    (theSum, ) = req.waitForReply(timeout)
    assert theSum == 6

    # But it is OK to register with less args to get default
    sigSlot.registerSlot(funcTwoArgDef, numArgs=2)
    req = sigSlot.request("", "funcTwoArgDef", 2, 3)
    (theSum, ) = req.waitForReply(timeout)
    assert theSum == 47  # 2 + 3 + 42

    # Still, we must not miss to sent required arguments
    req = sigSlot.request("", "funcTwoArgDef", 1)
    with pytest.raises(RuntimeError):
        req.waitForReply(timeout)

    #
    # Register function with default keyword only argument.
    #
    # Automatic number of argument detection excludes the default
    sigSlot.registerSlot(funcKwargs)
    req = sigSlot.request("", "funcKwargs")
    (length, a) = req.waitForReply(timeout)
    assert length == 0
    assert a == 77

    # Can specify whatever number of args, but keyword only arg is fix:
    sigSlot.registerSlot(funcKwargs, "funcKwargs1", 1)
    req = sigSlot.request("", "funcKwargs1", -77)
    (length, a) = req.waitForReply(timeout)
    assert length == 1
    assert a == 77

    sigSlot.registerSlot(funcKwargs, "funcKwargs2", 2)
    req = sigSlot.request("", "funcKwargs2", 111, 222)
    (length, a) = req.waitForReply(timeout)
    assert length == 2
    assert a == 77


def test_sigslot_register_method(eventLoopFixt):

    timeout = 2000  # timeout in ms

    sigSlot = SignalSlotableWithSlots("sigSlot" + str(uuid.uuid4()))
    sigSlot.start()

    #
    # Ordinary registration - register under <method>.__name__
    #
    sigSlot.registerSlot(sigSlot.slotOneArg)
    sigSlot.request("", "slotOneArg", 1).waitForReply(timeout)

    # Register under defined name
    sigSlot.registerSlot(sigSlot.slotOneArg, "slotOneArg1")
    req = sigSlot.request("", "slotOneArg1", "no_matter")
    req.waitForReply(timeout)

    #
    # Register method with variable number of arguments.
    #
    # Automatic number of argument detection gives zero:
    sigSlot.registerSlot(sigSlot.slotVarArgs)
    req = sigSlot.request("", "slotVarArgs")
    (nArg, ) = req.waitForReply(timeout)
    assert nArg == 0
    # NOTE: We require STRICT match for number of arguments.
    #       It's in underlying C++ code as well.
    req = sigSlot.request("", "slotVarArgs", "extra_arg")
    with pytest.raises(RuntimeError) as excinfo:
        req.waitForReply(timeout)
        assert "mismatched number of args" in str(excinfo.value)

    # We can register '*args' function with fixed number
    sigSlot.registerSlot(sigSlot.slotVarArgs, "slotVarArgs1", 2)
    req = sigSlot.request("", "slotVarArgs1", "no", "matter")
    (nArg, ) = req.waitForReply(timeout)
    assert 2 == nArg

    # Then we must not miss to sent argument
    req = sigSlot.request("", "slotVarArgs1", "not_enough")
    with pytest.raises(RuntimeError):
        req.waitForReply(timeout)

    #
    # Register method with default argument.
    #
    # Automatic number of argument detection includes the default
    # (cannot change that for backward compatibility...).
    sigSlot.registerSlot(sigSlot.slotTwoArgDef, "slotTwoArgDef3")
    req = sigSlot.request("", "slotTwoArgDef3", 1, 2, 3)
    (theSum, ) = req.waitForReply(timeout)
    assert 6 == theSum

    # But it is OK to register with less args to get default
    sigSlot.registerSlot(sigSlot.slotTwoArgDef, numArgs=2)
    req = sigSlot.request("", "slotTwoArgDef", 1, 2)
    (theSum, ) = req.waitForReply(timeout)
    assert 45 == theSum  # 1 + 2 + 42

    # Still, we must not miss to sent required arguments
    req = sigSlot.request("", "slotTwoArgDef", 1)
    with pytest.raises(RuntimeError):
        req.waitForReply(timeout)

    #
    # Register method with default keyword only argument.
    #
    # Automatic number of argument detection excludes the default
    sigSlot.registerSlot(sigSlot.slotKwargs)
    req = sigSlot.request("", "slotKwargs")
    (length, a) = req.waitForReply(timeout)
    assert 0 == length
    assert 77 == a

    # Can specify whatever number of args, but keyword only arg is fix:
    sigSlot.registerSlot(sigSlot.slotKwargs, "slotKwargs1", 1)
    req = sigSlot.request("", "slotKwargs1", -77)
    (length, a) = req.waitForReply(timeout)
    assert 1 == length
    assert 77 == a

    sigSlot.registerSlot(sigSlot.slotKwargs, "slotKwargs2", 2)
    req = sigSlot.request("", "slotKwargs2", 111, 222)
    (length, a) = req.waitForReply(timeout)
    assert 2 == length
    assert 77 == a


def test_sigslot_register_lambda(eventLoopFixt):

    timeout = 2000  # timeout in ms

    sigSlot = SignalSlotableWithSlots("sigSlot" + str(uuid.uuid4()))
    sigSlot.start()

    # Lambdas have a (not so unique) __name__, taken by default.
    # Auto detection of num(arguments) works, default arg don't harm.
    sigSlot.registerSlot(lambda a, b, c, d=55: sigSlot.reply(a, b, c, d))
    req = sigSlot.request("", "<lambda>", 1, 2, 3, 5)
    ll, m, n, o = req.waitForReply(timeout)
    assert 1 == ll
    assert 2 == m
    assert 3 == n
    assert 5 == o

    # Make actually use of lambda's default argument
    sigSlot.registerSlot(lambda a, b, c, d=55: sigSlot.reply(a, b, c, d),
                         "lam", 3)
    req = sigSlot.request("", "lam", 8, 13, 21)
    ll, m, n, o = req.waitForReply(timeout)
    assert 8 == ll
    assert 13 == m
    assert 21 == n
    assert 55 == o

    # Test that 'partial' works for slots - they have no __name__
    def forPartial(a, b):
        sigSlot.reply(a, b)

    part = partial(forPartial, 42)  # 'default' for a
    # Need to define slotName and numArgs, auto detection fails
    sigSlot.registerSlot(part, "partial", 1)
    req = sigSlot.request("", "partial", 43)
    a, b = req.waitForReply(timeout)
    assert 42 == a
    assert 43 == b


def test_sigslot_unique_id(eventLoopFixt):

    oneId = "one" + str(uuid.uuid4())
    twoId = "two" + str(uuid.uuid4())
    one = SignalSlotable(oneId)
    two = SignalSlotable(twoId, Hash(), 30, Hash('type', 'sigslot'))
    with does_not_raise():
        one.start()
        two.start()
    assert one.getInstanceId() == oneId
    assert two.getInstanceId() == twoId
    one_info = one.getInstanceInfo()
    two_info = two.getInstanceInfo()
    assert one_info['type'] == "unknown"
    assert two_info['type'] == "sigslot"
    assert one_info['heartbeatInterval'] == 10
    assert two_info['heartbeatInterval'] == 30
    avail = one.getAvailableInstances()
    assert isinstance(avail, Hash) is True
    assert avail[oneId + '.countdown'] == 10
    assert avail[twoId + '.countdown'] == 30
    assert avail[oneId + '.instanceInfo.type'] == 'unknown'
    assert avail[twoId + '.instanceInfo.type'] == 'sigslot'
    assert (avail[oneId + '.instanceInfo.karaboVersion'] ==
            avail[twoId + '.instanceInfo.karaboVersion'])
    one_again = SignalSlotable(oneId)
    with pytest.raises(Exception):
        one_again.start()


def test_sigslot_request(eventLoopFixt):

    bobId = "bob" + str(uuid.uuid4())
    bob = SignalSlotable(bobId)
    bob.start()
    called = -1

    def slot0():
        nonlocal called
        called = 0
        bob.reply()
    bob.registerSlot(slot0)

    def slot1(i):
        nonlocal called
        called = i
        bob.reply(i)
    bob.registerSlot(slot1)

    def slot2(i, j):
        nonlocal called
        called = i + j
        bob.reply(i, j)
    bob.registerSlot(slot2)

    def slot3(i, j, k):
        nonlocal called
        called = i + j + k
        bob.reply(i, j, k)
    bob.registerSlot(slot3)

    def slot4(i, j, k, n):  # flake does not like l: 'ambiguous var. name'
        nonlocal called
        called = i + j + k + n
        bob.reply(i, j, k, n)
    bob.registerSlot(slot4)

    aliceId = "alice" + str(uuid.uuid4())
    alice = SignalSlotable(aliceId)
    alice.start()

    handled = False

    def wait_for_handled():
        max_count = 2000
        while not handled and max_count > 0:
            time.sleep(0.01)
            max_count -= 1

    def handle0():
        nonlocal handled
        handled = True

    req = alice.request(bobId, "slot0")
    req.receiveAsync0(handle0)
    wait_for_handled()
    assert handled is True
    assert called == 0
    # test again via return-argument-generic receiveAsync
    handled = False
    req = alice.request(bobId, "slot0")
    req.receiveAsync(handle0)
    wait_for_handled()
    assert 0 == called

    # Test one argument
    handled = False
    i = 0

    def handle1(ii):
        nonlocal handled, i
        handled = True
        i = ii

    req = alice.request(bobId, "slot1", 5)
    req.receiveAsync1(handle1)
    wait_for_handled()
    assert handled is True
    assert 5 == called
    assert 5 == i

    # test again via return-argument-generic receiveAsync
    handled = False
    i = 0
    called = -1
    req = alice.request(bobId, "slot1", 5)
    req.receiveAsync(handle1)
    wait_for_handled()
    assert handled is True
    assert 5 == called
    assert 5 == i

    # test 2 arguments slot
    handled = False
    called == -1
    i, j = 0, 0

    def handle2(ii, jj):
        nonlocal handled, i, j
        handled = True
        i, j = ii, jj

    req = alice.request(bobId, "slot2", 5, 7)
    req.receiveAsync2(handle2)
    wait_for_handled()
    assert 12 == called
    assert 5 == i
    assert 7 == j
    # test again via return-argument-generic receiveAsync
    handled = False
    i = j = 0
    called = -1
    req = alice.request(bobId, "slot2", 5, 7)
    req.receiveAsync(handle2)
    wait_for_handled()
    assert 12 == called
    assert 5 == i
    assert 7 == j

    # Test three arguments
    handled = False
    called = -1
    i = j = k = 0

    def handle3(ii, jj, kk):
        nonlocal handled, i, j, k
        handled = True
        i, j, k = ii, jj, kk

    req = alice.request(bobId, "slot3", 5, 7, 11)
    req.receiveAsync3(handle3)
    wait_for_handled()
    assert handled is True
    assert 23 == called
    assert 5 == i
    assert 7 == j
    assert 11 == k
    # test again via return-argument-generic receiveAsync
    handled = False
    i = j = k = 0
    called = -1
    req = alice.request(bobId, "slot3", 5, 7, 11)
    req.receiveAsync(handle3)
    wait_for_handled()
    assert 23 == called
    assert 5 == i
    assert 7 == j
    assert 11 == k

    # Test four arguments
    handled = False
    called = -1
    i = j = k = n = 0  # flake does not like variable named 'l'

    def handle4(ii, jj, kk, nn):
        nonlocal handled, i, j, k, n
        handled = True
        i, j, k, n = ii, jj, kk, nn

    req = alice.request(bobId, "slot4", 5, 7, 11, 13)
    req.receiveAsync4(handle4)
    wait_for_handled()
    assert 36 == called
    assert 5 == i
    assert 7 == j
    assert 11 == k
    assert 13 == n
    # test again via return-argument-generic receiveAsync
    handled = False
    i = j = k = n = 0
    called = -1
    req = alice.request(bobId, "slot4", 5, 7, 11, 13)
    req.receiveAsync(handle4)
    wait_for_handled()
    assert 36 == called
    assert 5 == i
    assert 7 == j
    assert 11 == k
    assert 13 == n

    # Test error handling
    called = -1

    def slotError():
        nonlocal called
        called = 42
        # Exception in Python code
        raise RuntimeError("What's the universe and the rest?")

    bob.registerSlot(slotError)

    def slotErrorCpp():
        nonlocal called
        called = 43
        # Exception in Karabo C++ code
        h = Hash()
        h["non_existing_key"]

    bob.registerSlot(slotErrorCpp)

    handled = False
    errorMsg = ""
    detailsMsg = ""

    def handleError(msg, details):
        nonlocal errorMsg, detailsMsg, handled
        errorMsg = msg
        detailsMsg = details
        handled = True

    # Slot raising a pure Python exception
    req = alice.request(bobId, "slotError")
    req.receiveAsync(lambda: None, handleError)
    wait_for_handled()
    assert handled is True
    assert 42 == called

    # This also tests a bit the exception handling:
    # errorMsg has the main message,
    # detailsMsg contains C++ and Python traces.
    # print(f'\n*** errorMsg "{errorMsg}"\n*** detailsMsg "{detailsMsg}"')
    msgWhoFailed = "Remote Exception from bob"
    msgSlotFailure = "Error in slot 'slotError' from alice"
    msgRuntimeError = "What's the universe and the rest?"

    assert msgWhoFailed in errorMsg
    assert msgSlotFailure in errorMsg
    assert msgRuntimeError in errorMsg
    assert "Exception Type....:" not in errorMsg  # also C++ trace

    assert msgWhoFailed in detailsMsg
    assert msgSlotFailure in detailsMsg
    assert msgRuntimeError in detailsMsg
    # C++ trace:
    assert "Exception with trace (listed from inner to outer):" in detailsMsg
    assert "Exception Type....:" in detailsMsg
    # Parts of Python trace
    assert "Traceback (most recent call last):" in detailsMsg
    assert __file__ in detailsMsg   # raising is here in this file
    assert "raise RuntimeError" in detailsMsg

    # Check how exception from underlying C++ looks like
    handled = False
    errorMsg = ""
    detailsMsg = ""
    called = -1

    req = alice.request(bobId, "slotErrorCpp")
    req.receiveAsync(lambda: None, handleError)
    wait_for_handled()
    assert handled is True
    assert 43 == called
    # msgWhoFailed = "Remote Exception from bob" as before
    msgSlotFailure = "Error in slot 'slotErrorCpp' from alice"
    msgKeyFailure = "Key 'non_existing_key' does not exist"
    assert msgWhoFailed in errorMsg
    assert msgSlotFailure in errorMsg
    assert msgKeyFailure in errorMsg
    assert "Exception Type....:" in errorMsg  # also C++ trace

    assert msgWhoFailed in detailsMsg
    assert msgSlotFailure in detailsMsg
    assert msgKeyFailure in errorMsg

    # C++ trace:
    assert "Exception with trace (listed from inner to outer):" in detailsMsg
    assert "Exception Type....:  Python Exception" in detailsMsg
    # Parts of Python trace
    assert "Traceback (most recent call last):" in detailsMsg
    assert __file__ in detailsMsg  # raising is here in this file
    assert 'h["non_existing_key"]' in detailsMsg  # the bad line

    # Test timeout handling
    handled = False
    errorMsg = ""
    detailsMsg = ""
    called = -1

    req = alice.request("non/existing/id", "slotNoMatter")
    req.receiveAsync(lambda: None, handleError, timeoutMs=1)
    wait_for_handled()

    assert "Timeout of asynchronous request" in errorMsg
    assert "Timeout of asynchronous request" in detailsMsg
    assert "Exception Type....:  Timeout Exception" in detailsMsg

    # Forbidden number of return values
    req = alice.request("doesnt/matter/id", "slotNoMatter")
    try:
        req.receiveAsync(lambda: None, numCallbackArgs=5)
    except RuntimeError as e:
        assert "Detected/specified 5 (> 4) arguments" in str(e)
    else:
        assert False, "Did not catch RuntimeError"

    # So far we registered local funcs and lambda, now check
    # member funcs and objects with __call__
    # 1) Object with ordinary __call__ member
    handled = False
    called = -1
    i = 0

    class Handler1:
        def __call__(self, ii):
            nonlocal handled, i
            handled, i = True, ii

    handle1 = Handler1()
    req = alice.request(bobId, "slot1", 5)
    # Number of arguments is properly detected:
    req.receiveAsync(handle1)
    wait_for_handled()
    assert handled is True
    assert 5 == i
    assert 5 == called

    # 2) Object with static __call__ member
    handled = False
    called = -1
    i = 0

    class Handler1a:
        @staticmethod
        def __call__(ii):
            nonlocal handled, i
            handled, i = True, ii

    handle1 = Handler1a()
    req = alice.request(bobId, "slot1", 6)
    # Also here number of arguments is properly detected:
    req.receiveAsync(handle1)
    wait_for_handled()
    assert handled is True
    assert 6 == called
    assert 6 == i

    # 3) Just a method of an object
    handled = False
    called = -1
    i = 0

    req = alice.request(bobId, "slot1", 7)
    # re-use __call__ member of object from above
    req.receiveAsync(handle1.__call__)
    wait_for_handled()
    assert handled is True
    assert 7 == called
    assert 7 == i

    # 4) Bad handler with a non-callable __call__ attribute
    class HandlerBad:
        def __init__(self):
            self.__call__ = None

    handleBad = HandlerBad()
    req = alice.request("doesnt/matter/id", "slotNoMatter")
    with pytest.raises(RuntimeError):
        req.receiveAsync(handleBad)
    called = -1  # reset for next test

    # wait for reply ...
    assert called == -1
    to = 2000  # timeout in ms
    res = alice.request(bobId, "slot0").waitForReply(to)
    assert len(res) == 0
    assert called == 0

    res = alice.request(bobId, "slot1", 5).waitForReply(to)
    assert len(res) == 1
    assert res[0] == 5
    assert called == 5

    req = alice.request(bobId, "slot2", 5, 7)
    res = req.waitForReply(to)
    assert len(res) == 2
    assert res[0] == 5
    assert res[1] == 7
    assert called == 12

    req = alice.request(bobId, "slot3", 5, 7, 11)
    res = req.waitForReply(to)
    assert len(res) == 3
    assert res[0] == 5
    assert res[1] == 7
    assert res[2] == 11
    assert called == 23

    req = alice.request(bobId, "slot4", 5, 7, 11, 13)
    res = req.waitForReply(to)
    assert len(res) == 4
    assert res[0] == 5
    assert res[1] == 7
    assert res[2] == 11
    assert res[3] == 13
    assert called == 36

    # Now error caused by raising Python exception
    req = alice.request(bobId, "slotError")
    with pytest.raises(RuntimeError) as excinfo:
        res = req.waitForReply(to)

    # The error string contains both, user friendly and detailed
    # message, separated by "\nDETAILS: "
    exceptStrs = str(excinfo.value).split("\nDETAILS: ", 1)

    assert len(exceptStrs) == 2, exceptStrs
    (friendlyMsg, details) = exceptStrs
    msgWhoFailed = "Remote Exception from bob"
    msgSlotFailure = "Error in slot 'slotError' from alice"
    msgRuntimeError = ("RuntimeError: "
                       "What's the universe and the rest?")
    assert msgWhoFailed in friendlyMsg
    assert msgSlotFailure in friendlyMsg
    assert msgRuntimeError in friendlyMsg
    assert msgWhoFailed in details, str(exceptStrs)
    assert msgSlotFailure in details
    assert msgRuntimeError in details
    # C++ trace:
    assert "Exception with trace (listed from inner to outer):" in details
    assert "Exception Type....:" in details
    # Parts of Python trace
    assert "Traceback (most recent call last):" in details
    assert __file__ in details  # raising is in this file
    assert "raise RuntimeError" in details

    # Now error caused by throwing C++ exception in C++ called from Py
    req = alice.request(bobId, "slotErrorCpp")
    with pytest.raises(RuntimeError) as excinfo:
        res = req.waitForReply(to)

    # Again, friendly and detailed message, separated by
    # "\nDETAILS: "
    exceptStrs = str(excinfo.value).split("\nDETAILS: ", 1)
    assert len(exceptStrs) == 2, exceptStrs
    (friendlyMsg, details) = exceptStrs
    msgWhoFailed = "Remote Exception from bob"
    msgSlotFailure = "Error in slot 'slotErrorCpp' from alice"
    msgKeyFailure = "Key 'non_existing_key' does not exist"
    assert msgWhoFailed in friendlyMsg
    assert msgSlotFailure in friendlyMsg
    assert msgKeyFailure in friendlyMsg

    assert msgWhoFailed in details
    assert msgSlotFailure in details
    assert msgKeyFailure in details

    # C++ trace:
    assert "Exception with trace (listed from inner to outer):" in details
    assert "Exception Type....:  Python Exception" in details
    # Parts of Python trace
    assert "Traceback (most recent call last):" in details
    assert __file__ in details   # raising is this file
    assert 'h["non_existing_key"]' in details  # the bad line

    # Now timeout error
    req = alice.request("not/an/instance", "slotDoesNotMatter")
    with pytest.raises(TimeoutError):
        req.waitForReply(2)  # ms - short timeout intended

    del req


def test_sigslot_requestNoWait(eventLoopFixt):
    greeter = SignalSlotable("greeter")
    responder = SignalSlotable("responder")

    def slotA(value):
        responder.reply(value * 2)

    responder.registerSlot(slotA)

    replyCalled = repliedValue = None

    def slotReplyOfA(replyValue):
        nonlocal replyCalled, repliedValue
        repliedValue = replyValue
        replyCalled = True

    greeter.registerSlot(slotReplyOfA)

    greeter.start()
    responder.start()

    greeter.requestNoWait("responder", "slotA", "slotReplyOfA", 21)

    def wait_for_replyCalled():
        max_count = 2000
        while replyCalled is None and max_count > 0:
            time.sleep(0.01)
            max_count -= 1

    wait_for_replyCalled()

    assert replyCalled is True
    assert repliedValue == 42


def test_sigslot_asyncreply(eventLoopFixt):

    bobId = "bob" + str(uuid.uuid4())
    bob = SignalSlotable(bobId)
    bob.start()
    called = -1

    def asyncSlot0():
        aReply = bob.createAsyncReply()

        def replyer():
            nonlocal called
            called = 0
            aReply()

        # Call replyer on event loop after 100 ms
        EventLoop.post(replyer, 0.1)

    bob.registerSlot(asyncSlot0)

    def asyncSlot1(i):
        aReply = bob.createAsyncReply()

        def replyer():
            nonlocal called
            called = i
            aReply(i)

        # Call replyer on event loop after 10 ms
        EventLoop.post(replyer, 0.01)

    bob.registerSlot(asyncSlot1)

    def asyncSlot2(i, j):
        aReply = bob.createAsyncReply()

        def replyer():
            nonlocal called
            called = i + j
            aReply(i, j)

        # Call replyer on event loop
        EventLoop.post(replyer)

    bob.registerSlot(asyncSlot2)

    def asyncSlot3(i, j, k):
        aReply = bob.createAsyncReply()

        def replyer():
            nonlocal called
            called = i + j + k
            aReply(i, j, k)

        # Call replyer on event loop
        EventLoop.post(replyer)

    bob.registerSlot(asyncSlot3)

    def asyncSlot4(i, j, k, n):
        aReply = bob.createAsyncReply()

        def replyer():
            nonlocal called
            called = i + j + k + n
            aReply(i, j, k, n)

        # Call replyer on event loop
        EventLoop.post(replyer)

    bob.registerSlot(asyncSlot4)

    def asyncSlotError(errorMessage, details):
        aReply = bob.createAsyncReply()

        def replyer():
            aReply.error(errorMessage, details)

        # Call replyer on event loop
        EventLoop.post(replyer)

    bob.registerSlot(asyncSlotError)

    # setup other SignalSlotable to call the slots
    aliceId = "alice" + str(uuid.uuid4())
    alice = SignalSlotable(aliceId)
    alice.start()

    assert called == -1
    to = 2000  # timeout in ms
    req = alice.request(bobId, "asyncSlot0")
    res = req.waitForReply(to)
    assert 0 == len(res)
    assert 0 == called

    called = -1
    req = alice.request(bobId, "asyncSlot1", 5)
    res = req.waitForReply(to)
    assert 1 == len(res)
    assert 5 == res[0]
    assert 5 == called

    called = -1
    req = alice.request(bobId, "asyncSlot2", 5, 7)
    res = req.waitForReply(to)
    assert 2 == len(res)
    assert 5 == res[0]
    assert 7 == res[1]
    assert 12 == called

    called = -1
    req = alice.request(bobId, "asyncSlot3", 5, 7, 11)
    res = req.waitForReply(to)
    assert 3 == len(res)
    assert 5 == res[0]
    assert 7 == res[1]
    assert 11 == res[2]
    assert 23 == called

    called = -1
    req = alice.request(bobId, "asyncSlot4", 5, 7, 11, 13)
    res = req.waitForReply(to)
    assert 4 == len(res)
    assert 5 == res[0]
    assert 7 == res[1]
    assert 11 == res[2]
    assert 13 == res[3]
    assert 36 == called
    del res

    msg = "Bad things happen!"
    details = "No details nor trace!"
    req = alice.request(bobId, "asyncSlotError", msg, details)
    with pytest.raises(RuntimeError) as excinfo:
        req.waitForReply(to)

    # Again, friendly and detailed message, separated by
    # "\nDETAILS: "
    exceptStrs = str(excinfo.value).split("\nDETAILS: ", 1)
    assert 2 == len(exceptStrs), exceptStrs
    (friendlyMsg, detailsMsg) = exceptStrs
    msgWhoFailed = "Remote Exception from bob"
    assert msgWhoFailed in friendlyMsg
    assert msg in friendlyMsg
    assert msgWhoFailed in detailsMsg
    assert details in detailsMsg
    del req


def test_sigslot_asyncConnect(eventLoopFixt):
    signaler = SignalSlotable("signalInstance")
    signaler.registerSignal("signal", int)
    signaler.registerSignal("not_connected_signal", str)
    signaler.start()

    slotter = SignalSlotable("slotInstance")

    slotCalled = False
    inSlot = -10

    def slotFunc(i):
        nonlocal slotCalled, inSlot
        inSlot += i
        slotCalled = True

    slotter.registerSlot(slotFunc, "slot")
    slotter.start()

    # First test successful connectAsync
    failureMsg = None
    failureDetails = None

    def connectCallback(failMsg, failDetails):
        nonlocal failureMsg, failureDetails
        failureMsg = failMsg
        failureDetails = failDetails

    slotter.asyncConnect("signalInstance", "signal", "slot", connectCallback)

    def wait_for_callback():
        max_count = 4000  # 20 sec timeout
        while failureMsg is None and failureDetails is None and max_count > 0:
            time.sleep(0.005)
            max_count -= 1

    wait_for_callback()

    assert failureMsg is not None  # callback called
    assert failureMsg == ""        # connect succeeded

    signaler.emit("signal", 52)

    def wait_for_slotCalled(max_count=4000):  # 20 sec timeout
        count = max_count
        while not slotCalled and count > 0:
            time.sleep(0.005)
            count -= 1

    wait_for_slotCalled()

    assert slotCalled
    assert 42 == inSlot

    ##################################################
    # Test failure handling

    # Non-existing signal gives failure - but in Karabo 3 only since
    # signalInstance is in same process, otherwise just subscribes to broker
    failureMsg = failureDetails = None

    slotter.asyncConnect("signalInstance", "NOT_A_signal", "slot",
                         connectCallback)
    wait_for_callback()

    assert failureMsg is not None  # callback called
    assert "no signal 'NOT_A_signal'" in failureMsg
    # Details contain the same, but also C++ exception etails
    assert "no signal 'NOT_A_signal'" in failureDetails
    assert "Exception Type....:  SignalSlot Exception" in failureDetails

    # Non-existing slot gives failure
    failureMsg = failureDetails = None

    slotter.asyncConnect("signalInstance", "signal", "NOT_A_slot",
                         connectCallback)
    wait_for_callback()

    assert failureMsg is not None  # callback called
    assert "no slot 'NOT_A_slot'" in failureMsg
    # Details contain the same, but also C++ exception etails
    assert "no slot 'NOT_A_slot'" in failureDetails
    assert "Exception Type....:  SignalSlot Exception" in failureDetails

    #################################################################
    # Now asyncronously disconnect
    failureMsg = failureDetails = None

    slotter.asyncDisconnect("signalInstance", "signal", "slot",
                            connectCallback)
    wait_for_callback()

    assert failureMsg is not None  # callback called
    assert failureMsg == ""        # successfully disconnected

    slotCalled = False
    signaler.emit("signal", 77)

    # wait only a bit (10 means 50 ms) for something not coming
    wait_for_slotCalled(10)
    assert not slotCalled

    #################################################################
    # Finally a failing asyncDisconnect
    failureMsg = failureDetails = None

    slotter.asyncDisconnect("signalInstance", "signal", "not_a_slot",
                            connectCallback)
    wait_for_callback()

    assert failureMsg is not None  # callback called
    assert "no slot 'not_a_slot'" in failureMsg
    assert "no slot 'not_a_slot'" in failureDetails
    assert "Exception Type....:  SignalSlot Exception" in failureDetails



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.

from karabo.bound import Hash, HashNode, Types


def test_hashNode():
    # Test "getKey", "setValue", "getValue", "getValueAs",
    #      "setAttribute", "getAttribute", "getAttributeAs",
    #      "hasAttribute", "setAttributes", "getAttributes",
    #      "copyAttributes", "getType", "setType"

    # 'getValueAs' for HashNode is testing also the other methods:
    # `getAs`, `getAttributeAs` for Hash and HashAttributesNode.
    # Only one conversion version is used.
    # One exception to that is Hash's `getAs` for few types.

    # BOOL
    # Boolean is something special which can be converted
    # to any other type
    h = Hash("a.b.c", True)
    assert isinstance(h, Hash)
    node = h.getNode("a.b.c")
    assert isinstance(node, HashNode)

    # Test "getKey"
    assert node.getKey() == 'c'

    # Test "getValue"
    assert node.getValue() is True

    # Test "getType"
    assert node.getType() == Types.BOOL

    # Test "setType"
    node.setType("VECTOR_BOOL")
    assert node.getType() == Types.VECTOR_BOOL
    assert node.getValue() == [True]
    node.setType(Types.VECTOR_INT32)
    assert node.getType() == Types.VECTOR_INT32
    assert node.getValue() == [1]

    # Test "setValue"
    node.setValue([1, 2, 3])
    assert node.getValue() == [1, 2, 3]
    node.setValue(2.7818281828)
    assert node.getValue() == 2.7818281828
    assert node.getType() == Types.DOUBLE

    # Test "getValueAs"
    node.setValue(True)
    assert node.getValueAs(Types.CHAR) == '1'
    assert node.getValueAs("INT8") == 1
    # "getValueAs" uses default conversion tested in
    # "test_hash_attributes_node.py"
    # .... skipped tests
    assert node.getValueAs(Types.DOUBLE) == 1.0
    assert node.getValueAs("COMPLEX_FLOAT") == (1 + 0j)
    assert node.getValueAs("COMPLEX_DOUBLE") == (1 + 0J)
    assert node.getValueAs(Types.STRING) == '1'
    assert node.getValueAs(Types.VECTOR_BOOL) == [True]
    assert node.getValueAs(Types.VECTOR_INT16) == [1]
    # .... skipped tests ...
    assert node.getValueAs(Types.VECTOR_FLOAT) == [1.0]
    assert node.getValueAs(Types.VECTOR_DOUBLE) == [1.0]
    assert node.getValueAs(Types.VECTOR_COMPLEX_FLOAT) == [(1 + 0j)]
    assert node.getValueAs(Types.VECTOR_COMPLEX_DOUBLE) == [(1 + 0j)]
    # Call is not supported ...
    # FIXME: VECTOR_STRING not supported in both cases

    # INT32
    # Special value 1 or 0 can be converted like boolean to any type...
    h = Hash('a.b.c', 1)
    n = h.getNode('a.b.c')
    assert n.getKey() == 'c'
    assert n.getValue() == 1
    assert n.getType() == Types.INT32
    assert n.getValueAs(Types.BOOL) is True
    assert n.getValueAs(Types.CHAR) == '1'
    assert n.getValueAs(Types.INT8) == 1
    # ... skipped tests ...
    assert n.getValueAs(Types.UINT64) == 1
    assert n.getValueAs(Types.FLOAT) == 1.0
    assert n.getValueAs(Types.DOUBLE) == 1.0
    assert n.getValueAs(Types.COMPLEX_FLOAT) == (1 + 0j)
    assert n.getValueAs(Types.COMPLEX_DOUBLE) == (1 + 0j)
    assert n.getValueAs(Types.STRING) == '1'
    assert n.getValueAs(Types.VECTOR_BOOL) == [True]
    # assert n.getValueAs(Types.VECTOR_CHAR) == ['1']
    # assert n.getValueAs(Types.VECTOR_INT8) == [1]
    # assert n.getValueAs(Types.VECTOR_UINT8) == [1]
    assert n.getValueAs(Types.VECTOR_INT16) == [1]
    # ... skipped tests
    assert n.getValueAs(Types.VECTOR_UINT64) == [1]
    assert n.getValueAs(Types.VECTOR_FLOAT) == [1.0]
    assert n.getValueAs(Types.VECTOR_DOUBLE) == [1.0]
    assert n.getValueAs(Types.VECTOR_COMPLEX_FLOAT) == [(1 + 0j)]
    assert n.getValueAs(Types.VECTOR_COMPLEX_DOUBLE) == [(1 + 0j)]
    # Call is not supported ...
    # assert n.getValueAs(Types.VECTOR_STRING) == ['1']

    h = Hash("a.b.c", 42)
    n = h.getNode("a.b.c")
    assert n.getValue() == 42
    assert n.getType() == Types.INT32
    assert n.getValueAs(Types.INT8) == 42
    # ... skipped tests
    assert n.getValueAs(Types.DOUBLE) == 42
    assert n.getValueAs(Types.COMPLEX_FLOAT) == (42 + 0j)
    assert n.getValueAs(Types.COMPLEX_DOUBLE) == (42 + 0j)
    assert n.getValueAs(Types.STRING) == '42'
    # Base64 encoding
    # Since we use `karabo::util::StringTools.hh` for conversion
    # the following types are converted using Base64 encoding, so
    # we skip them here...
    # assert n.getValueAs(Types.VECTOR_INT8) == [42]
    # assert n.getValueAs(Types.VECTOR_UINT8) == [42]
    assert n.getValueAs(Types.VECTOR_INT16) == [42]
    # ... skipped tests
    assert n.getValueAs(Types.VECTOR_UINT64) == [42]
    assert n.getValueAs(Types.VECTOR_FLOAT) == [42.0]
    assert n.getValueAs(Types.VECTOR_DOUBLE) == [42.0]
    assert n.getValueAs("VECTOR_COMPLEX_FLOAT") == [(42 + 0j)]
    assert n.getValueAs("VECTOR_COMPLEX_DOUBLE") == [(42 + 0j)]
    # Call is not supported...
    # assert n.getValueAs(Types.VECTOR_STRING) == '42'

    # Test "setAttribute"
    h = Hash("a", 42)
    h.setAttribute("a", "a", 15)
    node = h.getNode("a")
    assert isinstance(node, HashNode)

    # Test "getAttribute"
    assert node.getAttribute("a") == 15

    # Test "getAttributeAs"
    assert node.getAttributeAs("a", Types.INT8) == 15
    # ... skipped tests
    assert node.getAttributeAs("a", "DOUBLE") == 15
    assert node.getAttributeAs("a", Types.COMPLEX_FLOAT) == (15 + 0j)
    assert node.getAttributeAs("a", Types.COMPLEX_DOUBLE) == (15 + 0j)
    assert node.getAttributeAs("a", Types.STRING) == '15'
    # Base64 encoding
    # assert node.getAttributeAs("a", Types.VECTOR_INT8) == [15]
    # assert node.getAttributeAs("a", Types.VECTOR_UINT8) == [15]
    assert node.getAttributeAs("a", "VECTOR_INT16") == [15]
    # ... skipped tests
    assert node.getAttributeAs("a", Types.VECTOR_DOUBLE) == [15]
    assert node.getAttributeAs("a", Types.VECTOR_COMPLEX_FLOAT) == [
        (15 + 0j)]
    assert node.getAttributeAs("a", Types.VECTOR_COMPLEX_DOUBLE) == [
        (15 + 0j)]
    # Not supported
    # assert node.getAttributeAs("a", Types.VECTOR_STRING) == ['15']

    # Test "hasAttribute"
    assert node.hasAttribute("a")
    assert not node.hasAttribute("b")

    # Test "setAttributes", "getAttributes", "copyAttributes"
    h = Hash('a', 12)
    n1 = h.getNode('a')
    n1.setAttribute('a', 1)
    n1.setAttribute('b', True)
    n1.setAttribute('c', "test")
    n1.setAttribute('d', 3.141592)

    # Test "getAttributes"
    attrs = n1.getAttributes()
    assert attrs['a'] == 1
    assert attrs['d'] == 3.141592
    attrs['b'] = False
    assert not n1.getAttribute('b')

    h['b'] = "value"
    n2 = h.getNode('b')

    # Test "setAttributes"
    n2.setAttributes(attrs)  # copy
    n1.setAttribute('c', "newtest")
    assert n2.getAttribute('c') != "newtest"

    # Test "copyAttributes"
    attrs2 = n1.copyAttributes()
    attrs2['a'] = 2
    assert n1.getAttribute('a') != 2



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
# To change this template, choose Tools | Templates
# and open the template in the editor.

from karabo.bound import (
    BOOL_ELEMENT, FLOAT_ELEMENT, INT32_ELEMENT, INT64_ELEMENT, NODE_ELEMENT,
    STRING_ELEMENT, UINT32_ELEMENT, AssemblyRules, Hash, HashFilter,
    MetricPrefix, Unit, Validator)
from karabo.bound.decorators import (
    KARABO_CLASSINFO, KARABO_CONFIGURATION_BASE_CLASS)
from karabo.common.states import State


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("BaseY", "1.0")
class BaseY:
    def __init__(self, configuration):
        super().__init__()


@KARABO_CLASSINFO("P1Y", "1.0")
class P1Y(BaseY):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("a")
            .description("a").displayedName("a")
            .assignmentOptional().defaultValue("a value")
            .tags("CY,CY,NC,JS,KW,NC")
            .commit(),

            STRING_ELEMENT(expected).key("b")
            .tags("BH,CY")
            .displayedName("Example key 1")
            .description("Example key 1 description")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Air Condition")
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("c").alias(10)
            .tags("BH")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.METER).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(5)
            .init()
            .commit(),

            UINT32_ELEMENT(expected).key("d").alias(5.5)
            .tags("CY,JS")
            .displayedName("Example key 3")
            .description("Example key 3 description")
            .allowedStates(State.STARTED, State.STOPPED,
                           State.NORMAL)  # TODO check
            .minExc(10).maxExc(20).assignmentOptional().defaultValue(11)
            .reconfigurable()
            .commit(),

            FLOAT_ELEMENT(expected).key("e").alias("exampleAlias4")
            .tags("DB,NC,CY")
            .displayedName("Example key 4")
            .description("Example key 4 description")
            .options("1.1100000, -2.22, 5.55")
            .assignmentOptional()
            .defaultValue(1.11)
            .commit(),

            INT64_ELEMENT(expected).key("f").alias("exampleAlias5")
            .tags("LM,DB")
            .displayedName("Example key 5")
            .description("Example key 5 description")
            .assignmentOptional()
            .defaultValue(5)
            .commit(),
        )


@KARABO_CLASSINFO("P2Y", "1.0")
class P2Y(BaseY):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("x")
            .description("x")
            .displayedName("x")
            .assignmentOptional().defaultValue("a value")
            .tags("LM,BH")
            .commit(),

            STRING_ELEMENT(expected).key("y")
            .tags("CY")
            .displayedName("Example key 1")
            .description("Example key 1 description")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Radio")
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("z").alias(10)
            .tags("CY,LM,KW")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.AMPERE).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),
        )


@KARABO_CLASSINFO("P3Y", "1.0")
class P3Y(BaseY):
    def __init__(self, configuration):
        super().__init__(configuration)

    @staticmethod
    def expectedParameters(expected):
        (
            STRING_ELEMENT(expected).key("k")
            .description("k").displayedName("k")
            .assignmentOptional().defaultValue("k value")
            .tags("LM")
            .commit(),

            STRING_ELEMENT(expected).key("l")
            .tags("CY")
            .displayedName("l").description("l")
            .options("Radio,Air Condition,Navigation", ",")
            .assignmentOptional().defaultValue("Navigation")
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("m").alias(10)
            .tags("CY,DB,JE,BP,MK,PG,BF")
            .displayedName("Example key 2")
            .description("Example key 2 description")
            .options("5, 25, 10")
            .minInc(5).maxInc(25).unit(Unit.METER).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(25)
            .init()
            .commit(),
        )


@KARABO_CONFIGURATION_BASE_CLASS
@KARABO_CLASSINFO("GraphicsRenderer2Y", "1.0")
class GraphicsRenderer2Y:
    def __init__(self, configuration):
        super().__init__()

    @staticmethod
    def expectedParameters(expected):
        (
            BOOL_ELEMENT(expected).key("antiAlias")
            .tags("NC")
            .displayedName("Use Anti-Aliasing")
            .description("You may switch of for speed")
            .assignmentOptional().defaultValue(True)
            .init().expertAccess()
            .commit(),

            STRING_ELEMENT(expected).key("color")
            .tags("KW")
            .displayedName("Color")
            .description("The default color for any shape")
            .assignmentOptional().defaultValue("red")
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected).key("bold")
            .tags("LM")
            .displayedName("Bold").description("Toggles bold painting")
            .assignmentOptional().defaultValue(False)
            .reconfigurable()
            .commit(),

            STRING_ELEMENT(expected).key("shapes")
            .tags("DB")
            .assignmentOptional().defaultValue("rectangle")
            .commit(),

            NODE_ELEMENT(expected).key("circle")
            .tags("JS")
            .displayedName("Circle").description("A circle")
            .commit(),

            FLOAT_ELEMENT(expected).key("circle.radius")
            .description("The radius of the circle")
            .displayedName("Radius")
            .tags("NC,KW")
            .minExc(0).maxExc(100).unit(Unit.METER).metricPrefix(
                MetricPrefix.MILLI)
            .assignmentOptional().defaultValue(10)
            .init().commit(),

            NODE_ELEMENT(expected).key("rectangle")
            .tags("BH, KW , CY")
            .displayedName("Rectangle").description("A rectangle")
            .commit(),

            FLOAT_ELEMENT(expected).key("rectangle.b")
            .tags("JS")
            .description("Rectangle side - b").displayedName("Side B")
            .assignmentOptional().defaultValue(10)
            .init().commit(),

            FLOAT_ELEMENT(expected).key("rectangle.c")
            .tags("LM,JS")
            .description("Rectangle side - c").displayedName("Side C")
            .assignmentOptional().defaultValue(10)
            .init()
            .commit(),

            NODE_ELEMENT(expected).key("triangle")
            .displayedName("triangle")
            .description("A triangle (Node element containing no "
                         "other elements)")
            .commit(),

            NODE_ELEMENT(expected).key("letter")
            .displayedName("Letter").description("Letter")
            .appendParametersOf(P1Y)
            .commit(),

        )


def test_hash_filter():
    schema = GraphicsRenderer2Y.getSchema(
        "GraphicsRenderer2Y", rules=AssemblyRules())
    validator = Validator()
    _, _, config = validator.validate(schema, Hash())

    result = HashFilter.byTag(schema, config, "KW;KW,BH", ",;")

    assert ("antiAlias" in result) is False
    assert ("color" in result) is True
    assert ("bold" in result) is False
    assert ("shapes" in result) is False
    assert ("rectangle" in result) is True
    assert ("rectangle.b" in result) is True
    assert ("rectangle.c" in result) is True
    assert ("letter" in result) is True
    assert ("letter.a" in result) is True
    assert ("letter.b" in result) is True
    assert ("letter.c" in result) is True
    assert ("letter.d" in result) is False
    assert ("letter.e" in result) is False
    assert ("letter.f" in result) is False

    result = HashFilter.byTag(schema, config, "JS", ",;")

    assert ("antiAlias" in result) is False
    assert ("bold" in result) is False
    assert ("shapes" in result) is False
    assert ("rectangle" in result) is True
    assert ("rectangle.b" in result) is True
    assert ("rectangle.c" in result) is True
    assert ("letter" in result) is True
    assert ("letter.a" in result) is True
    assert ("letter.b" in result) is False
    assert ("letter.c" in result) is False
    assert ("letter.d" in result) is True
    assert ("letter.e" in result) is False
    assert ("letter.f" in result) is False

    result = HashFilter.byTag(schema, config, "NC,LM", ",;")

    assert ("antiAlias" in result) is True
    assert ("color" in result) is False
    assert ("bold" in result) is True
    assert ("shapes" in result) is False
    assert ("rectangle" in result) is True
    assert ("rectangle.b" in result) is False
    assert ("rectangle.c" in result) is True
    assert ("letter" in result) is True
    assert ("letter.a" in result) is True
    assert ("letter.b" in result) is False
    assert ("letter.c" in result) is False
    assert ("letter.d" in result) is False
    assert ("letter.e" in result) is True
    assert ("letter.f" in result) is True

    result = HashFilter.byTag(schema, config, "CY", ",;")

    assert ("antiAlias" in result) is False
    assert ("color" in result) is False
    assert ("bold" in result) is False
    assert ("shapes" in result) is False
    assert ("rectangle" in result) is True
    assert ("rectangle.b" in result) is True
    assert ("rectangle.c" in result) is True
    assert ("letter" in result) is True
    assert ("letter.a" in result) is True
    assert ("letter.b" in result) is True
    assert ("letter.c" in result) is False
    assert ("letter.d" in result) is True
    assert ("letter.e" in result) is True
    assert ("letter.f" in result) is False

    result = HashFilter.byTag(schema, config, "BF", ",;")

    assert ("antiAlias" in result) is False
    assert ("color" in result) is False
    assert ("bold" in result) is False
    assert ("shapes" in result) is False
    assert ("rectangle" in result) is False
    assert ("rectangle.b" in result) is False
    assert ("rectangle.c" in result) is False
    assert ("letter" in result) is False
    assert ("letter.a" in result) is False
    assert ("letter.b" in result) is False
    assert ("letter.c" in result) is False
    assert ("letter.d" in result) is False
    assert ("letter.e" in result) is False
    assert ("letter.f" in result) is False

    # no "WP76" tag
    result = HashFilter.byTag(schema, config, "WP76", ",;")

    assert ("antiAlias" in result) is False
    assert ("color" in result) is False
    assert ("bold" in result) is False
    assert ("shapes" in result) is False
    assert ("rectangle" in result) is False
    assert ("rectangle.b" in result) is False
    assert ("rectangle.c" in result) is False
    assert ("letter" in result) is False
    assert ("letter.a" in result) is False
    assert ("letter.b" in result) is False
    assert ("letter.c" in result) is False
    assert ("letter.d" in result) is False
    assert ("letter.e" in result) is False
    assert ("letter.f" in result) is False



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
""" This is a test device, written in the bound API, for the
API cross test
"""

import numpy

from karabo.bound import (
    BOOL_ELEMENT, DOUBLE_ELEMENT, IMAGEDATA_ELEMENT, INPUT_CHANNEL,
    INT32_ELEMENT, KARABO_CLASSINFO, NDARRAY_ELEMENT, NODE_ELEMENT,
    OUTPUT_CHANNEL, SLOT_ELEMENT, STRING_ELEMENT, TABLE_ELEMENT,
    VECTOR_STRING_ELEMENT, AlarmCondition, ChannelMetaData, Encoding,
    Epochstamp, Hash, ImageData, MetricPrefix, PythonDevice, Schema, State,
    TimeId, Timestamp, Types, Unit)


@KARABO_CLASSINFO("TestDevice", "1.5")
class TestDevice(PythonDevice):
    @staticmethod
    def expectedParameters(expected):
        tableSchema = Schema()
        (
            DOUBLE_ELEMENT(tableSchema).key("e")
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.KILO)
            .assignmentOptional()
            .defaultValue(0.0)
            .commit(),

            STRING_ELEMENT(tableSchema).key("s")
            .assignmentOptional()
            .defaultValue("")
            .commit()
        )
        pipeSchema = Schema()
        (
            # the default type of a NDARRAY elements and of IMAGEDATA
            # are now UNKNOWN. Here are left unspecified to test that
            # the integration tests do not require this definition.
            NDARRAY_ELEMENT(pipeSchema).key("ndarray")
            .dtype(Types.FLOAT)
            .shape("10")
            .commit(),

            IMAGEDATA_ELEMENT(pipeSchema).key("image")
            .setDimensions("50,50")
            .setType(Types.UINT16)
            .setEncoding(Encoding.GRAY)
            .commit(),

            # note, this will never be set by `send`
            # it is to prove that a raw channel input can manage
            # data mismatching the schema and an normal channel
            # does not fail in case of mismatching schema.
            DOUBLE_ELEMENT(pipeSchema).key("e")
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.KILO)
            .assignmentOptional()
            .noDefaultValue()
            .commit(),

            STRING_ELEMENT(pipeSchema).key("s")
            .assignmentOptional()
            .noDefaultValue()
            .commit()
        )
        (
            DOUBLE_ELEMENT(expected).key("a")
            .minExc(22).maxExc(33).minInc(11).maxInc(23)
            .displayedName("parameter a")
            .alias("something")
            .tags("bla,blub")
            .description("a's description")
            .allowedStates(State.INIT, State.UNKNOWN)
            .unit(Unit.AMPERE)
            .metricPrefix(MetricPrefix.MILLI)
            .expertAccess()
            .assignmentOptional()
            .defaultValue(22.5)
            .reconfigurable()
            .commit(),

            BOOL_ELEMENT(expected).key("eosReceived")
            .displayedName("EOS Received")
            .assignmentOptional()
            .defaultValue(False)
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("readonly")
            .displayedName("Readonly")
            .description("A readonly parameter")
            .readOnly()
            .initialValue(2)
            .commit(),

            NODE_ELEMENT(expected).key("node")
            .commit(),

            DOUBLE_ELEMENT(expected).key("node.b")
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.KILO)
            .options("33,44,55,100")
            .assignmentOptional()
            .defaultValue(33)
            .commit(),

            STRING_ELEMENT(expected).key("middlelayerDevice")
            .assignmentMandatory()
            .reconfigurable()
            .commit(),

            INT32_ELEMENT(expected).key("maxSizeSchema")
            .readOnly()
            .initialValue(0)
            .commit(),

            SLOT_ELEMENT(expected).key("setA")
            .commit(),

            SLOT_ELEMENT(expected).key("backfire")
            .commit(),

            SLOT_ELEMENT(expected).key("injectSchema")
            .commit(),

            SLOT_ELEMENT(expected).key("send")
            .commit(),

            SLOT_ELEMENT(expected).key("sendMultipleHashes")
            .commit(),

            SLOT_ELEMENT(expected).key("end")
            .commit(),

            SLOT_ELEMENT(expected).key("compareSchema")
            .commit(),

            TABLE_ELEMENT(expected).key("table")
            .displayedName("bla")
            .setNodeSchema(tableSchema)
            .assignmentOptional()
            .defaultValue([Hash("e", 5, "s", "hallo")])
            .reconfigurable()
            .commit(),

            NDARRAY_ELEMENT(expected).key("ndarray")
            .shape("3,2")
            .unit(Unit.METER)
            .metricPrefix(MetricPrefix.KILO)
            .dtype("FLOAT")
            .commit(),

            OUTPUT_CHANNEL(expected).key("output1")
            .dataSchema(tableSchema)
            .commit(),

            OUTPUT_CHANNEL(expected).key("output2")
            .dataSchema(pipeSchema)
            .commit(),

            INPUT_CHANNEL(expected).key("input")
            .commit(),

            VECTOR_STRING_ELEMENT(expected).key("interfaces")
            .displayedName("Interfaces")
            .assignmentOptional().defaultValue(
                ["Motor", "Camera", "Processor", "DeviceInstantiator"])
            .commit(),

            INPUT_CHANNEL(expected).key("imageInput")
            .commit(),

            STRING_ELEMENT(expected).key("imagePath")
            .displayedName("Image Path")
            .assignmentOptional().defaultValue("data.image")
            .expertAccess()
            .init()
            .commit(),

            STRING_ELEMENT(expected).key("ndarrayPath")
            .displayedName("NDArray Path")
            .assignmentOptional().defaultValue("data.array")
            .expertAccess()
            .init()
            .commit(),

            INT32_ELEMENT(expected).key("imagesReceived")
            .readOnly()
            .initialValue(0)
            .commit(),

            INT32_ELEMENT(expected).key("ndarraysReceived")
            .readOnly()
            .initialValue(0)
            .commit(),

        )

    def __init__(self, configuration):
        super().__init__(configuration)
        self.registerInitialFunction(self.initialize)
        self.registerSlot(self.setA)
        self.registerSlot(self.backfire)
        self.registerSlot(self.injectSchema)
        self.registerSlot(self.send)
        self.registerSlot(self.sendMultipleHashes)
        self.registerSlot(self.end)
        self.registerSlot(self.compareSchema)
        self.KARABO_ON_DATA("input", self.onData)
        self.KARABO_ON_EOS("input", self.onEndOfStream)
        self.KARABO_ON_DATA("imageInput", self.onImageData)
        self.word_no = 1

    def initialize(self):
        # HACK/FIXME:
        # We trigger a temporary (15 s) connection to our counterpart and fill
        # the remote's cache already here in initialisation.
        # Otherwise this will block in the slot calls where it is used - and
        # block forever since one cannot synchronously call back to a slot
        # caller due to thte way ordering per sender is implemented.
        mdlDevice = self.get("middlelayerDevice")
        self.remote().getDeviceSchema(mdlDevice)
        self.remote().get(mdlDevice)

    def setA(self):
        ts = Timestamp(Epochstamp("20090901T135522"), TimeId(0))
        self.set("a", 22.7, ts)
        ts = Timestamp(Epochstamp("20160617T135522"), TimeId(0))
        self.set("node.b", 100, ts)
        self.set("ndarray", numpy.array([[1, 2, 3], [4, 5, 6]]))

    def backfire(self):
        remote = self.remote()
        instance_id = self.get("middlelayerDevice")

        state = remote.get(instance_id, "state")
        # 'isinstance' instead of 'is' allows inheritance in future
        if not isinstance(state, State):
            raise RuntimeError("Middle layer 'state' property is of type",
                               type(state).__name__)
        alarm = remote.get(instance_id, "alarmCondition")
        if not isinstance(alarm, AlarmCondition):
            raise RuntimeError("Middle layer 'alarmCondition' property "
                               "is of type", type(alarm).__name__)
        # FIXME:
        # In principle we should use set and execute here, i.e. no 'NoWait'.
        # But, within a slot call, it does not work to call back synchronously
        # to the caller of the slot: The reply we expect here will not be
        # processed because we block here in backfire and messages are executed
        # in order of arrival per sender...
        # Could be fixed by posting the synchronous calls out of this thread
        # and using an AsyncReply - which does not yet exist for bound Python.
        remote.setNoWait(instance_id, "value", 99)
        remote.executeNoWait(instance_id, "slot")

    def injectSchema(self):
        schema = Schema()

        (
            STRING_ELEMENT(schema).key(f"word{self.word_no}")
            .displayedName(f"Word #{self.word_no}")
            .description("The word")
            .assignmentOptional().defaultValue("Hello")
            .reconfigurable()
            .commit(),

            NODE_ELEMENT(schema).key("injectedNode")
            .commit(),

            DOUBLE_ELEMENT(schema)
            .key(f"injectedNode.number{self.word_no}")
            .assignmentOptional()
            .noDefaultValue()
            .commit()
        )
        self.updateSchema(schema)
        self.set(f"injectedNode.number{self.word_no}", self.word_no)
        self.word_no += 1

    def send(self):
        self.writeChannel("output1", Hash("e", 5., "s", "hallo"))
        node = Hash("e", 5., "s", "hallo")
        arr = numpy.full((10), 42., dtype=numpy.float32)
        node.set("ndarray", arr)
        imArr = numpy.full((50, 50), 42, dtype=numpy.uint16)
        node.set("image", ImageData(imArr, encoding=Encoding.GRAY))
        self.writeChannel("output2", node)

    def sendMultipleHashes(self):
        channel = self._sigslot.getOutputChannel("output1")

        for i in range(10):
            sourceName = f"{self.getInstanceId()}:output{i}"
            timestamp = self.getActualTimestamp()
            meta = ChannelMetaData(sourceName, timestamp)
            channel.write(Hash("e", float(i), "s", "hallo"), meta)
        channel.update()

    def end(self):
        self.signalEndOfStream("output1")
        self.signalEndOfStream("output2")

    def onData(self, data, metaData):
        self.set("a", data.get("number"))

    def onEndOfStream(self, channel):
        self.set("eosReceived", True)

    def checkData(self, data, update, dataPrefix, klass):
        prop_path = self[f"{dataPrefix}Path"]
        if data.has(prop_path) and isinstance(data[prop_path], klass):
            count = self[f"{dataPrefix}sReceived"] + 1
            update.set(f"{dataPrefix}sReceived", count)
        else:
            update.set("status", "unexpected data type"
                       f" '{type(data[prop_path])}'"
                       f" instead of '{str(klass)}'"
                       f" for path '{prop_path}'")

    def onImageData(self, data, metaData):
        h = Hash()
        self.checkData(data, h, "image", ImageData)
        self.checkData(data, h, "ndarray", numpy.ndarray)
        self.set(h)

    def compareSchema(self):
        """This function tries to get the maxSize of a middlelayer device
        """
        instance_id = self.get("middlelayerDevice")
        remote = self.remote()
        schema = remote.getDeviceSchema(instance_id)
        max_size = schema.getMaxSize("vectorMaxSize")
        self.set("maxSizeSchema", max_size)
        self.reply(max_size)



# This file is part of Karabo.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# Karabo is free software: you can redistribute it and/or modify it under
# the terms of the MPL-2 Mozilla Public License.
#
# You should have received a copy of the MPL-2 Public License along with
# Karabo. If not, see <https://www.mozilla.org/en-US/MPL/2.0/>.
#
# Karabo is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.
from karabo.bound import Hash


def test_find_1():
    h = Hash("a.b.c1.d", 1)

    node = h.find("a.b.c1.d")
    assert node is not None
    assert node.getKey() == "d"
    assert node.getValue() == 1

    node = h.find("a.b.c1.f")
    assert node is None


def test_find_2():
    h = Hash("a.b.c", "1")

    node = h.find("a.b.c")
    assert node is not None
    node.setValue(2)
    assert h.get("a.b.c") == 2

    node = h.find("a.b.c", '/')
    assert node is None
